# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements OS shell Job functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals

from . import _async
from . import _job


class LambdaJob(_job.Job):
  """Represents an lambda function as a Job.
  """

  def __init__(self, fn, job_id, *args, **kwargs):
    """Initializes an instance of a Job.

    Args:
      fn: the lambda function to execute asyncronously
      job_id: an optional ID for the job. If None, a UUID will be generated.
    """
    super(LambdaJob, self).__init__(job_id)
    self._future = _async.async.executor.submit(fn, *args, **kwargs)

  def __repr__(self):
    """Returns a representation for the job for showing in the notebook.
    """
    return 'Job %s %s' % (self._job_id, self.state)

  #TODO: ShellJob, once we need it, should inherit on LambdaJob:
  #      import subprocess
  #      LambdaJob(subprocess.check_output, id, command_line, shell=True)



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements GCP Job functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals

import datalab.context
from . import _job


class GCPJob(_job.Job):
  """Represents a BigQuery Job.
  """

  def __init__(self, job_id, context):
    """Initializes an instance of a Job.

    Args:
      job_id: the BigQuery job ID corresponding to this job.
      context: a Context object providing project_id and credentials.
    """
    super(GCPJob, self).__init__(job_id)
    if context is None:
      context = datalab.context.Context.default()
    self._context = context
    self._api = self._create_api(context)

  def _create_api(self, context):
    raise Exception('_create_api must be defined in a derived class')

  def __repr__(self):
    """Returns a representation for the job for showing in the notebook.
    """
    return 'Job %s/%s %s' % (self._context.project_id, self._job_id, self.state)




# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - BigQuery SQL Dialect"""
from __future__ import absolute_import


class Dialect(object):
  """
  Represents the default BigQuery SQL dialect
  """
  _global_dialect = None

  def __init__(self, bq_dialect):
    self._global_dialect = bq_dialect

  @property
  def bq_dialect(self):
    """Retrieves the value of the bq_dialect property.

    Returns:
      The default BigQuery SQL dialect
    """
    return self._global_dialect

  def set_bq_dialect(self, bq_dialect):
    """ Set the default BigQuery SQL dialect"""
    if bq_dialect in ['legacy', 'standard']:
      self._global_dialect = bq_dialect

  @staticmethod
  def default():
    """Retrieves the default BigQuery SQL dialect, creating it if necessary.

    Returns:
      An initialized and shared instance of a Dialect object.
    """
    if Dialect._global_dialect is None:
      Dialect._global_dialect = Dialect('legacy')
    return Dialect._global_dialect



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implementation of command parsing and handling within magics."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

try:
  import IPython
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import argparse
import shlex


class CommandParser(argparse.ArgumentParser):
  """An argument parser to parse commands in line/cell magic declarations. """

  def __init__(self, *args, **kwargs):
    """Initializes an instance of a CommandParser. """
    super(CommandParser, self).__init__(*args, **kwargs)
    self._subcommands = None

  @staticmethod
  def create(name):
    """Creates a CommandParser for a specific magic. """
    return CommandParser(prog=name)

  def exit(self, status=0, message=None):
    """Overridden exit method to stop parsing without calling sys.exit(). """
    raise Exception(message)

  def format_usage(self):
    """Overridden usage generator to use the full help message. """
    return self.format_help()

  @staticmethod
  def create_args(line, namespace):
    """ Expand any meta-variable references in the argument list. """
    args = []
    # Using shlex.split handles quotes args and escape characters.
    for arg in shlex.split(line):
      if not arg:
         continue
      if arg[0] == '$':
        var_name = arg[1:]
        if var_name in namespace:
          args.append((namespace[var_name]))
        else:
          raise Exception('Undefined variable referenced in command line: %s' % arg)
      else:
        args.append(arg)
    return args

  def parse(self, line, namespace=None):
    """Parses a line into a dictionary of arguments, expanding meta-variables from a namespace. """
    try:
      if namespace is None:
        ipy = IPython.get_ipython()
        namespace = ipy.user_ns
      args = CommandParser.create_args(line, namespace)
      return self.parse_args(args)
    except Exception as e:
      print(str(e))
      return None

  def subcommand(self, name, help):
    """Creates a parser for a sub-command. """
    if self._subcommands is None:
      self._subcommands = self.add_subparsers(help='commands')
    return self._subcommands.add_parser(name, description=help, help=help)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements BigQuery query job functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str

from . import _job
from . import _query_results_table


class QueryJob(_job.Job):
  """ Represents a BigQuery Query Job. """

  def __init__(self, job_id, table_name, sql, context):
    """  Initializes a QueryJob object.

    Args:
      job_id: the ID of the query job.
      table_name: the name of the table where the query results will be stored.
      sql: the SQL statement that was executed for the query.
      context: the Context object providing project_id and credentials that was used
          when executing the query.
    """
    super(QueryJob, self).__init__(job_id, context)
    self._sql = sql
    self._table = _query_results_table.QueryResultsTable(table_name, context, self,
                                                         is_temporary=True)
    self._bytes_processed = None
    self._cache_hit = None
    self._total_rows = None

  @property
  def bytes_processed(self):
    """ The number of bytes processed, or None if the job is not complete. """
    return self._bytes_processed

  @property
  def total_rows(self):
    """ The total number of rows in the result, or None if not complete. """
    return self._total_rows

  @property
  def cache_hit(self):
    """ Whether the query results were obtained from the cache or not, or None if not complete. """
    return self._cache_hit

  @property
  def sql(self):
    """ The SQL statement that was executed for the query. """
    return self._sql

  def wait(self, timeout=None):
    """ Wait for the job to complete, or a timeout to happen.

      This is more efficient than the version in the base Job class, in that we can
      use a call that blocks for the poll duration rather than a sleep. That means we
      shouldn't block unnecessarily long and can also poll less.

    Args:
      timeout: how long to wait (in seconds) before giving up; default None which means no timeout.

    Returns:
      The QueryJob
    """
    poll = 30
    while not self._is_complete:
      try:
        query_result = self._api.jobs_query_results(self._job_id,
                                                    project_id=self._context.project_id,
                                                    page_size=0,
                                                    timeout=poll * 1000)
      except Exception as e:
        raise e
      if query_result['jobComplete']:
        if 'totalBytesProcessed' in query_result:
          self._bytes_processed = int(query_result['totalBytesProcessed'])
        self._cache_hit = query_result.get('cacheHit', None)
        if 'totalRows' in query_result:
          self._total_rows = int(query_result['totalRows'])
        break

      if timeout is not None:
        timeout -= poll
        if timeout <= 0:
          break

    self._refresh_state()
    return self

  @property
  def results(self):
    """ Get the table used for the results of the query. If the query is incomplete, this blocks.

    Raises:
      Exception if we timed out waiting for results or the query failed.
    """
    self.wait()
    if self.failed:
      raise Exception('Query failed: %s' % str(self.errors))
    return self._table




# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Query BigQuery API."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

import datalab.context
import datalab.data
import datalab.utils

from . import _api
from . import _federated_table
from . import _query_job
from . import _sampling
from . import _udf
from . import _utils


class Query(object):
  """Represents a Query object that encapsulates a BigQuery SQL query.

  This object can be used to execute SQL queries and retrieve results.
  """

  @staticmethod
  def sampling_query(sql, context, fields=None, count=5, sampling=None, udfs=None,
                     data_sources=None):
    """Returns a sampling Query for the SQL object.

    Args:
      sql: the SQL statement (string) or Query object to sample.
      context: a Context object providing project_id and credentials.
      fields: an optional list of field names to retrieve.
      count: an optional count of rows to retrieve which is used if a specific
          sampling is not specified.
      sampling: an optional sampling strategy to apply to the table.
      udfs: array of UDFs referenced in the SQL.
      data_sources: dictionary of federated (external) tables referenced in the SQL.
    Returns:
      A Query object for sampling the table.
    """
    return Query(_sampling.Sampling.sampling_query(sql, fields, count, sampling), context=context,
                 udfs=udfs, data_sources=data_sources)

  def __init__(self, sql, context=None, values=None, udfs=None, data_sources=None, **kwargs):
    """Initializes an instance of a Query object.
       Note that either values or kwargs may be used, but not both.

    Args:
      sql: the BigQuery SQL query string to execute, or a SqlStatement object. The latter will
          have any variable references replaced before being associated with the Query (i.e.
          once constructed the SQL associated with a Query is static).

          It is possible to have variable references in a query string too provided the variables
          are passed as keyword arguments to this constructor.

      context: an optional Context object providing project_id and credentials. If a specific
          project id or credentials are unspecified, the default ones configured at the global
          level are used.
      values: a dictionary used to expand variables if passed a SqlStatement or a string with
          variable references.
      udfs: array of UDFs referenced in the SQL.
      data_sources: dictionary of federated (external) tables referenced in the SQL.
      kwargs: arguments to use when expanding the variables if passed a SqlStatement
          or a string with variable references.

    Raises:
      Exception if expansion of any variables failed.
      """
    if context is None:
      context = datalab.context.Context.default()
    self._context = context
    self._api = _api.Api(context)
    self._data_sources = data_sources
    self._udfs = udfs

    if data_sources is None:
      data_sources = {}

    self._results = None
    self._code = None
    self._imports = []
    if values is None:
      values = kwargs

    self._sql = datalab.data.SqlModule.expand(sql, values)

    # We need to take care not to include the same UDF code twice so we use sets.
    udfs = set(udfs if udfs else [])
    for value in list(values.values()):
      if isinstance(value, _udf.UDF):
        udfs.add(value)
    included_udfs = set([])

    tokens = datalab.data.tokenize(self._sql)
    udf_dict = {udf.name: udf for udf in udfs}

    for i, token in enumerate(tokens):
      # Find the preceding and following non-whitespace tokens
      prior = i - 1
      while prior >= 0 and tokens[prior].isspace():
        prior -= 1
      if prior < 0:
        continue
      next = i + 1
      while next < len(tokens) and tokens[next].isspace():
        next += 1
      if next >= len(tokens):
        continue

      uprior = tokens[prior].upper()
      if uprior != 'FROM' and uprior != 'JOIN':
        continue

      # Check for external tables.
      if tokens[next] not in "[('\"":
        if token not in data_sources:
          if values and token in values:
            if isinstance(values[token], _federated_table.FederatedTable):
              data_sources[token] = values[token]

      # Now check for UDF calls.
      if uprior != 'FROM' or tokens[next] != '(':
        continue

      # We have a 'FROM token (' sequence.

      if token in udf_dict:
        udf = udf_dict[token]
        if token not in included_udfs:
          included_udfs.add(token)
          if self._code is None:
            self._code = []
          self._code.append(udf.code)
          if udf.imports:
            self._imports.extend(udf.imports)

        fields = ', '.join([f[0] for f in udf._outputs])
        tokens[i] = '(SELECT %s FROM %s' % (fields, token)

        # Find the closing parenthesis and add the additional one now needed.
        num_paren = 0
        j = i + 1
        while j < len(tokens):
          if tokens[j] == '(':
            num_paren += 1
          elif tokens[j] == ')':
            num_paren -= 1
            if num_paren == 0:
              tokens[j] = '))'
              break
          j += 1

    self._external_tables = None
    if len(data_sources):
      self._external_tables = {}
      for name, table in list(data_sources.items()):
        if table.schema is None:
          raise Exception('Referenced external table %s has no known schema' % name)
        self._external_tables[name] = table._to_query_json()

    self._sql = ''.join(tokens)

  def _repr_sql_(self):
    """Creates a SQL representation of this object.

    Returns:
      The SQL representation to use when embedding this object into other SQL.
    """
    return '(%s)' % self._sql

  def __str__(self):
    """Creates a string representation of this object.

    Returns:
      The string representation of this object (the unmodified SQL).
    """
    return self._sql

  def __repr__(self):
    """Creates a friendly representation of this object.

    Returns:
      The friendly representation of this object (the unmodified SQL).
    """
    return self._sql

  @property
  def sql(self):
    """ Get the SQL for the query. """
    return self._sql

  @property
  def scripts(self):
    """ Get the code for any Javascript UDFs used in the query. """
    return self._code

  def results(self, use_cache=True, dialect=None, billing_tier=None):
    """Retrieves table of results for the query. May block if the query must be executed first.

    Args:
      use_cache: whether to use cached results or not. Ignored if append is specified.
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A QueryResultsTable containing the result set.
    Raises:
      Exception if the query could not be executed or query response was malformed.
    """
    if not use_cache or (self._results is None):
      self.execute(use_cache=use_cache, dialect=dialect, billing_tier=billing_tier)
    return self._results.results

  def extract(self, storage_uris, format='csv', csv_delimiter=',', csv_header=True,
              compress=False, use_cache=True, dialect=None, billing_tier=None):
    """Exports the query results to GCS.

    Args:
      storage_uris: the destination URI(s). Can be a single URI or a list.
      format: the format to use for the exported data; one of 'csv', 'json', or 'avro'
          (default 'csv').
      csv_delimiter: for csv exports, the field delimiter to use (default ',').
      csv_header: for csv exports, whether to include an initial header line (default True).
      compress: whether to compress the data on export. Compression is not supported for
          AVRO format (default False).
      use_cache: whether to use cached results or not (default True).
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A Job object for the export Job if it was completed successfully; else None.
    Raises:
      An Exception if the query or extract failed.
    """
    return self.results(use_cache=use_cache, dialect=dialect,
                        billing_tier=billing_tier).extract(storage_uris, format=format,
                                                           csv_delimiter=csv_delimiter,
                                                           csv_header=csv_header,
                                                           compress=compress)

  @datalab.utils.async_method
  def extract_async(self, storage_uris, format='csv', csv_delimiter=',', csv_header=True,
                    compress=False, use_cache=True, dialect=None, billing_tier=None):
    """Exports the query results to GCS. Returns a Job immediately.

    Note that there are two jobs that may need to be run sequentially, one to run the query,
    and the second to extract the resulting table. These are wrapped by a single outer Job.

    If the query has already been executed and you would prefer to get a Job just for the
    extract, you can can call extract_async on the QueryResultsTable instead; i.e.:

        query.results().extract_async(...)

    Args:
      storage_uris: the destination URI(s). Can be a single URI or a list.
      format: the format to use for the exported data; one of 'csv', 'json', or 'avro'
          (default 'csv').
      csv_delimiter: for CSV exports, the field delimiter to use (default ',').
      csv_header: for CSV exports, whether to include an initial header line (default True).
      compress: whether to compress the data on export. Compression is not supported for
          AVRO format (default False).
      use_cache: whether to use cached results or not (default True).
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A Job for the combined (execute, extract) task that will in turn return the Job object for
      the completed extract task when done; else None.
    Raises:
      An Exception if the query failed.
    """
    return self.extract(storage_uris, format=format, csv_delimiter=csv_delimiter,
                        csv_header=csv_header, use_cache=use_cache, compress=compress,
                        dialect=dialect, billing_tier=billing_tier)

  def to_dataframe(self, start_row=0, max_rows=None, use_cache=True, dialect=None,
                   billing_tier=None):
    """ Exports the query results to a Pandas dataframe.

    Args:
      start_row: the row of the table at which to start the export (default 0).
      max_rows: an upper limit on the number of rows to export (default None).
      use_cache: whether to use cached results or not (default True).
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A Pandas dataframe containing the table data.
    """
    return self.results(use_cache=use_cache, dialect=dialect, billing_tier=billing_tier) \
        .to_dataframe(start_row=start_row, max_rows=max_rows)

  def to_file(self, path, format='csv', csv_delimiter=',', csv_header=True, use_cache=True,
              dialect=None, billing_tier=None):
    """Save the results to a local file in CSV format.

    Args:
      path: path on the local filesystem for the saved results.
      format: the format to use for the exported data; currently only 'csv' is supported.
      csv_delimiter: for CSV exports, the field delimiter to use. Defaults to ','
      csv_header: for CSV exports, whether to include an initial header line. Default true.
      use_cache: whether to use cached results or not.
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      The path to the local file.
    Raises:
      An Exception if the operation failed.
    """
    self.results(use_cache=use_cache, dialect=dialect, billing_tier=billing_tier) \
        .to_file(path, format=format, csv_delimiter=csv_delimiter, csv_header=csv_header)
    return path

  @datalab.utils.async_method
  def to_file_async(self, path, format='csv', csv_delimiter=',', csv_header=True, use_cache=True,
                    dialect=None, billing_tier=None):
    """Save the results to a local file in CSV format. Returns a Job immediately.

    Args:
      path: path on the local filesystem for the saved results.
      format: the format to use for the exported data; currently only 'csv' is supported.
      csv_delimiter: for CSV exports, the field delimiter to use. Defaults to ','
      csv_header: for CSV exports, whether to include an initial header line. Default true.
      use_cache: whether to use cached results or not.
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A Job for the save that returns the path to the local file on completion.
    Raises:
      An Exception if the operation failed.
    """
    return self.to_file(path, format=format, csv_delimiter=csv_delimiter, csv_header=csv_header,
                        use_cache=use_cache, dialect=dialect, billing_tier=billing_tier)

  def sample(self, count=5, fields=None, sampling=None, use_cache=True, dialect=None,
             billing_tier=None):
    """Retrieves a sampling of rows for the query.

    Args:
      count: an optional count of rows to retrieve which is used if a specific
          sampling is not specified (default 5).
      fields: the list of fields to sample (default None implies all).
      sampling: an optional sampling strategy to apply to the table.
      use_cache: whether to use cached results or not (default True).
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A QueryResultsTable containing a sampling of the result set.
    Raises:
      Exception if the query could not be executed or query response was malformed.
    """
    return Query.sampling_query(self._sql, self._context, count=count, fields=fields,
                                sampling=sampling, udfs=self._udfs,
                                data_sources=self._data_sources).results(use_cache=use_cache,
                                                                         dialect=dialect,
                                                                         billing_tier=billing_tier)

  def execute_dry_run(self, dialect=None, billing_tier=None):
    """Dry run a query, to check the validity of the query and return some useful statistics.

    Args:
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A dict with 'cacheHit' and 'totalBytesProcessed' fields.
    Raises:
      An exception if the query was malformed.
    """
    try:
      query_result = self._api.jobs_insert_query(self._sql, self._code, self._imports, dry_run=True,
                                                 table_definitions=self._external_tables, dialect=dialect,
                                                 billing_tier=billing_tier)
    except Exception as e:
      raise e
    return query_result['statistics']['query']

  def execute_async(self, table_name=None, table_mode='create', use_cache=True,
                    priority='interactive', allow_large_results=False, dialect=None,
                    billing_tier=None):
    """ Initiate the query and return a QueryJob.

    Args:
      table_name: the result table name as a string or TableName; if None (the default), then a
          temporary table will be used.
      table_mode: one of 'create', 'overwrite' or 'append'. If 'create' (the default), the request
          will fail if the table exists.
      use_cache: whether to use past query results or ignore cache. Has no effect if destination is
          specified (default True).
      priority:one of 'batch' or 'interactive' (default). 'interactive' jobs should be scheduled
          to run quickly but are subject to rate limits; 'batch' jobs could be delayed by as much
          as three hours but are not rate-limited.
      allow_large_results: whether to allow large results; i.e. compressed data over 100MB. This is
          slower and requires a table_name to be specified) (default False).
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A QueryJob.
    Raises:
      Exception if query could not be executed.
    """
    batch = priority == 'low'
    append = table_mode == 'append'
    overwrite = table_mode == 'overwrite'
    if table_name is not None:
      table_name = _utils.parse_table_name(table_name, self._api.project_id)

    try:
      query_result = self._api.jobs_insert_query(self._sql, self._code, self._imports,
                                                 table_name=table_name,
                                                 append=append,
                                                 overwrite=overwrite,
                                                 use_cache=use_cache,
                                                 batch=batch,
                                                 allow_large_results=allow_large_results,
                                                 table_definitions=self._external_tables,
                                                 dialect=dialect,
                                                 billing_tier=billing_tier)
    except Exception as e:
      raise e
    if 'jobReference' not in query_result:
      raise Exception('Unexpected response from server')

    job_id = query_result['jobReference']['jobId']
    if not table_name:
      try:
        destination = query_result['configuration']['query']['destinationTable']
        table_name = (destination['projectId'], destination['datasetId'], destination['tableId'])
      except KeyError:
        # The query was in error
        raise Exception(_utils.format_query_errors(query_result['status']['errors']))
    return _query_job.QueryJob(job_id, table_name, self._sql, context=self._context)

  def execute(self, table_name=None, table_mode='create', use_cache=True, priority='interactive',
              allow_large_results=False, dialect=None, billing_tier=None):
    """ Initiate the query, blocking until complete and then return the results.

    Args:
      table_name: the result table name as a string or TableName; if None (the default), then a
          temporary table will be used.
      table_mode: one of 'create', 'overwrite' or 'append'. If 'create' (the default), the request
          will fail if the table exists.
      use_cache: whether to use past query results or ignore cache. Has no effect if destination is
          specified (default True).
      priority:one of 'batch' or 'interactive' (default). 'interactive' jobs should be scheduled
          to run quickly but are subject to rate limits; 'batch' jobs could be delayed by as much
          as three hours but are not rate-limited.
      allow_large_results: whether to allow large results; i.e. compressed data over 100MB. This is
          slower and requires a table_name to be specified) (default False).
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      The QueryResultsTable for the query.
    Raises:
      Exception if query could not be executed.
    """
    job = self.execute_async(table_name=table_name, table_mode=table_mode, use_cache=use_cache,
                             priority=priority, allow_large_results=allow_large_results,
                             dialect=dialect, billing_tier=billing_tier)
    self._results = job.wait()
    return self._results

  def to_view(self, view_name):
    """ Create a View from this Query.

    Args:
      view_name: the name of the View either as a string or a 3-part tuple
          (projectid, datasetid, name).

    Returns:
      A View for the Query.
    """
    # Do the import here to avoid circular dependencies at top-level.
    from . import _view
    return _view.View(view_name, self._context).create(self._sql)




# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements HTTP client helper functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import str
from past.builtins import basestring
from builtins import object

import datetime
import json
import urllib.request, urllib.parse, urllib.error
import httplib2


# TODO(nikhilko): Start using the requests library instead.


class RequestException(Exception):

  def __init__(self, status, content):
    self.status = status
    self.content = content
    self.message = 'HTTP request failed'
    # Try extract a message from the body; swallow possible resulting ValueErrors and KeyErrors.
    try:
      error = json.loads(content)['error']
      if 'errors' in error:
        error = error['errors'][0]
      self.message += ': ' + error['message']
    except Exception:
      lines = content.split('\n') if isinstance(content, basestring) else []
      if lines:
        self.message += ': ' + lines[0]

  def __str__(self):
    return self.message


class Http(object):
  """A helper class for making HTTP requests.
  """

  def __init__(self):
    pass

  @staticmethod
  def request(url, args=None, data=None, headers=None, method=None,
              credentials=None, raw_response=False, stats=None):
    """Issues HTTP requests.

    Args:
      url: the URL to request.
      args: optional query string arguments.
      data: optional data to be sent within the request.
      headers: optional headers to include in the request.
      method: optional HTTP method to use. If unspecified this is inferred
          (GET or POST) based on the existence of request data.
      credentials: optional set of credentials to authorize the request.
      raw_response: whether the raw response content should be returned as-is.
      stats: an optional dictionary that, if provided, will be populated with some
          useful info about the request, like 'duration' in seconds and 'data_size' in
          bytes. These may be useful optimizing the access to rate-limited APIs.
    Returns:
      The parsed response object.
    Raises:
      Exception when the HTTP request fails or the response cannot be processed.
    """
    if headers is None:
      headers = {}

    headers['user-agent'] = 'GoogleCloudDataLab/1.0'
    # Add querystring to the URL if there are any arguments.
    if args is not None:
      qs = urllib.parse.urlencode(args)
      url = url + '?' + qs

    # Setup method to POST if unspecified, and appropriate request headers
    # if there is data to be sent within the request.
    if data is not None:
      if method is None:
        method = 'POST'

      if data != '':
        # If there is a content type specified, use it (and the data) as-is.
        # Otherwise, assume JSON, and serialize the data object.
        if 'Content-Type' not in headers:
          data = json.dumps(data)
          headers['Content-Type'] = 'application/json'
      headers['Content-Length'] = str(len(data))
    else:
      if method == 'POST':
        headers['Content-Length'] = '0'

    # If the method is still unset, i.e. it was unspecified, and there
    # was no data to be POSTed, then default to GET request.
    if method is None:
      method = 'GET'

    # Create an Http object to issue requests. Associate the credentials
    # with it if specified to perform authorization.
    #
    # TODO(nikhilko):
    # SSL cert validation seemingly fails, and workarounds are not amenable
    # to implementing in library code. So configure the Http object to skip
    # doing so, in the interim.
    http = httplib2.Http()
    http.disable_ssl_certificate_validation = True
    if credentials is not None:
      http = credentials.authorize(http)
    if stats is not None:
      stats['duration'] = datetime.datetime.utcnow()

    response = None
    try:
      response, content = http.request(url,
                                       method=method,
                                       body=data,
                                       headers=headers)
      if 200 <= response.status < 300:
        if raw_response:
          return content
        if type(content) == str:
          return json.loads(content)
        else:
          return json.loads(str(content, encoding='UTF-8'))
      else:
        raise RequestException(response.status, content)
    except ValueError:
      raise Exception('Failed to process HTTP response.')
    except httplib2.HttpLib2Error:
      raise Exception('Failed to send HTTP request.')
    finally:
      if stats is not None:
        stats['data_size'] = len(data)
        stats['status'] = response.status
        stats['duration'] = (datetime.datetime.utcnow() - stats['duration']).total_seconds()



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements BigQuery query job results table functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals

from . import _table


class QueryResultsTable(_table.Table):
  """ A subclass of Table specifically for Query results.

  The primary differences are the additional properties job_id and sql.
  """

  def __init__(self, name, context, job, is_temporary=False):
    """Initializes an instance of a Table object.

    Args:
      name: the name of the table either as a string or a 3-part tuple (projectid, datasetid, name).
      context: an optional Context object providing project_id and credentials. If a specific
        project id or credentials are unspecified, the default ones configured at the global
        level are used.
      job: the QueryJob associated with these results.
      is_temporary: if True, this is a short-lived table for intermediate results (default False).
    """
    super(QueryResultsTable, self).__init__(name, context)
    self._job = job
    self._is_temporary = is_temporary

  def __repr__(self):
    """Returns a representation for the dataset for showing in the notebook.
    """
    if self._is_temporary:
      return 'QueryResultsTable %s' % self.job_id
    else:
      return super(QueryResultsTable, self).__repr__()

  @property
  def job(self):
    """ The QueryJob object that caused the table to be populated. """
    return self._job

  @property
  def job_id(self):
    """ The ID of the query job that caused the table to be populated. """
    return self._job.id

  @property
  def sql(self):
    """ The SQL statement for the query that populated the table. """
    return self._job.sql

  @property
  def is_temporary(self):
    """ Whether this is a short-lived table or not. """
    return self._is_temporary



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Utility functions."""
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from builtins import str
from past.builtins import basestring

try:
  import IPython
  import IPython.core.display
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import json
import pandas
try:
  # Pandas profiling is not needed for build/test but will be in the container.
  import pandas_profiling
except ImportError:
  pass
import sys
import types
import yaml

import datalab.data
import datalab.bigquery
import datalab.storage
import datalab.utils

from . import _html


def notebook_environment():
  """ Get the IPython user namespace. """
  ipy = IPython.get_ipython()
  return ipy.user_ns


def get_notebook_item(name):
  """ Get an item from the IPython environment. """
  env = notebook_environment()
  return datalab.utils.get_item(env, name)


def render_list(data):
  return IPython.core.display.HTML(_html.HtmlBuilder.render_list(data))


def render_dictionary(data, headers=None):
  """ Return a dictionary list formatted as a HTML table.

  Args:
    data: the dictionary list
    headers: the keys in the dictionary to use as table columns, in order.
  """
  return IPython.core.display.HTML(_html.HtmlBuilder.render_table(data, headers))


def render_text(text, preformatted=False):
  """ Return text formatted as a HTML

  Args:
    text: the text to render
    preformatted: whether the text should be rendered as preformatted
  """
  return IPython.core.display.HTML(_html.HtmlBuilder.render_text(text, preformatted))


def get_field_list(fields, schema):
  """ Convert a field list spec into a real list of field names.

      For tables, we return only the top-level non-RECORD fields as Google charts
      can't handle nested data.
  """
  # If the fields weren't supplied get them from the schema.
  if isinstance(fields, list):
    return fields
  if isinstance(fields, basestring) and fields != '*':
    return fields.split(',')
  if not schema:
    return []
  return [f['name'] for f in schema._bq_schema if f['type'] != 'RECORD']


def _get_cols(fields, schema):
  """ Get column metadata for Google Charts based on field list and schema. """
  typemap = {
    'STRING': 'string',
    'INT64': 'number',
    'INTEGER': 'number',
    'FLOAT': 'number',
    'FLOAT64': 'number',
    'BOOL': 'boolean',
    'BOOLEAN': 'boolean',
    'DATE': 'date',
    'TIME': 'timeofday',
    'DATETIME': 'datetime',
    'TIMESTAMP': 'datetime'
  }
  cols = []
  for col in fields:
    if schema:
      f = schema[col]
      t = 'string' if f.mode == 'REPEATED' else typemap.get(f.data_type, 'string')
      cols.append({'id': f.name, 'label': f.name, 'type': t})
    else:
      # This will only happen if we had no rows to infer a schema from, so the type
      # is not really important, except that GCharts will choke if we pass such a schema
      # to a chart if it is string x string so we default to number.
      cols.append({'id': col, 'label': col, 'type': 'number'})
  return cols


def _get_data_from_empty_list(source, fields='*', first_row=0, count=-1, schema=None):
  """ Helper function for _get_data that handles empty lists. """
  fields = get_field_list(fields, schema)
  return {'cols': _get_cols(fields, schema), 'rows': []}, 0


def _get_data_from_list_of_dicts(source, fields='*', first_row=0, count=-1, schema=None):
  """ Helper function for _get_data that handles lists of dicts. """
  if schema is None:
    schema = datalab.bigquery.Schema.from_data(source)
  fields = get_field_list(fields, schema)
  gen = source[first_row:first_row + count] if count >= 0 else source
  rows = [{'c': [{'v': row[c]} if c in row else {} for c in fields]} for row in gen]
  return {'cols': _get_cols(fields, schema), 'rows': rows}, len(source)


def _get_data_from_list_of_lists(source, fields='*', first_row=0, count=-1, schema=None):
  """ Helper function for _get_data that handles lists of lists. """
  if schema is None:
    schema = datalab.bigquery.Schema.from_data(source)
  fields = get_field_list(fields, schema)
  gen = source[first_row:first_row + count] if count >= 0 else source
  cols = [schema.find(name) for name in fields]
  rows = [{'c': [{'v': row[i]} for i in cols]} for row in gen]
  return {'cols': _get_cols(fields, schema), 'rows': rows}, len(source)


def _get_data_from_dataframe(source, fields='*', first_row=0, count=-1, schema=None):
  """ Helper function for _get_data that handles Pandas DataFrames. """
  if schema is None:
    schema = datalab.bigquery.Schema.from_data(source)
  fields = get_field_list(fields, schema)
  rows = []
  if count < 0:
    count = len(source.index)
  df_slice = source.reset_index(drop=True)[first_row:first_row + count]
  for index, data_frame_row in df_slice.iterrows():
    row = data_frame_row.to_dict()
    for key in list(row.keys()):
      val = row[key]
      if isinstance(val, pandas.Timestamp):
        row[key] = val.to_pydatetime()

    rows.append({'c': [{'v': row[c]} if c in row else {} for c in fields]})
  cols = _get_cols(fields, schema)
  return {'cols': cols, 'rows': rows}, len(source)


def _get_data_from_table(source, fields='*', first_row=0, count=-1, schema=None):
  """ Helper function for _get_data that handles BQ Tables. """
  if not source.exists():
    return _get_data_from_empty_list(source, fields, first_row, count)
  if schema is None:
    schema = source.schema
  fields = get_field_list(fields, schema)
  gen = source.range(first_row, count) if count >= 0 else source
  rows = [{'c': [{'v': row[c]} if c in row else {} for c in fields]} for row in gen]
  return {'cols': _get_cols(fields, schema), 'rows': rows}, source.length


def get_data(source, fields='*', env=None, first_row=0, count=-1, schema=None):
  """ A utility function to get a subset of data from a Table, Query, Pandas dataframe or List.

  Args:
    source: the source of the data. Can be a Table, Pandas DataFrame, List of dictionaries or
        lists, or a string, in which case it is expected to be the name of a table in BQ.
    fields: a list of fields that we want to return as a list of strings, comma-separated string,
        or '*' for all.
    env: if the data source is a Query module, this is the set of variable overrides for
        parameterizing the Query.
    first_row: the index of the first row to return; default 0. Onl;y used if count is non-negative.
    count: the number or rows to return. If negative (the default), return all rows.
    schema: the schema of the data. Optional; if supplied this can be used to help do type-coercion.

  Returns:
    A tuple consisting of a dictionary and a count; the dictionary has two entries: 'cols'
    which is a list of column metadata entries for Google Charts, and 'rows' which is a list of
    lists of values. The count is the total number of rows in the source (independent of the
    first_row/count parameters).

  Raises:
    Exception if the request could not be fulfilled.
  """

  ipy = IPython.get_ipython()
  if env is None:
    env = {}
  env.update(ipy.user_ns)
  if isinstance(source, basestring):
    source = datalab.utils.get_item(ipy.user_ns, source, source)
    if isinstance(source, basestring):
      source = datalab.bigquery.Table(source)

  if isinstance(source, types.ModuleType) or isinstance(source, datalab.data.SqlStatement):
    source = datalab.bigquery.Query(source, values=env)

  if isinstance(source, list):
    if len(source) == 0:
      return _get_data_from_empty_list(source, fields, first_row, count, schema)
    elif isinstance(source[0], dict):
      return _get_data_from_list_of_dicts(source, fields, first_row, count, schema)
    elif isinstance(source[0], list):
      return _get_data_from_list_of_lists(source, fields, first_row, count, schema)
    else:
      raise Exception("To get tabular data from a list it must contain dictionaries or lists.")
  elif isinstance(source, pandas.DataFrame):
    return _get_data_from_dataframe(source, fields, first_row, count, schema)
  elif isinstance(source, datalab.bigquery.Query):
    return _get_data_from_table(source.results(), fields, first_row, count, schema)
  elif isinstance(source, datalab.bigquery.Table):
    return _get_data_from_table(source, fields, first_row, count, schema)
  else:
    raise Exception("Cannot chart %s; unsupported object type" % source)


def handle_magic_line(line, cell, parser, namespace=None):
  """ Helper function for handling magic command lines given a parser with handlers set. """
  args = parser.parse(line, namespace)
  if args:
    try:
      return args.func(vars(args), cell)
    except Exception as e:
      sys.stderr.write(str(e))
      sys.stderr.write('\n')
      sys.stderr.flush()
  return None


def expand_var(v, env):
  """ If v is a variable reference (for example: '$myvar'), replace it using the supplied
      env dictionary.

  Args:
    v: the variable to replace if needed.
    env: user supplied dictionary.

  Raises:
    Exception if v is a variable reference but it is not found in env.
  """
  if len(v) == 0:
    return v
  # Using len() and v[0] instead of startswith makes this Unicode-safe.
  if v[0] == '$':
    v = v[1:]
    if len(v) and v[0] != '$':
      if v in env:
        v = env[v]
      else:
        raise Exception('Cannot expand variable $%s' % v)
  return v


def replace_vars(config, env):
  """ Replace variable references in config using the supplied env dictionary.

  Args:
    config: the config to parse. Can be a tuple, list or dict.
    env: user supplied dictionary.

  Raises:
    Exception if any variable references are not found in env.
  """
  if isinstance(config, dict):
    for k, v in list(config.items()):
      if isinstance(v, dict) or isinstance(v, list) or isinstance(v, tuple):
        replace_vars(v, env)
      elif isinstance(v, basestring):
        config[k] = expand_var(v, env)
  elif isinstance(config, list):
    for i, v in enumerate(config):
      if isinstance(v, dict) or isinstance(v, list) or isinstance(v, tuple):
        replace_vars(v, env)
      elif isinstance(v, basestring):
        config[i] = expand_var(v, env)
  elif isinstance(config, tuple):
    # TODO(gram): figure out how to handle these if the tuple elements are scalar
    for v in config:
      if isinstance(v, dict) or isinstance(v, list) or isinstance(v, tuple):
        replace_vars(v, env)


def parse_config(config, env, as_dict=True):
  """ Parse a config from a magic cell body. This could be JSON or YAML. We turn it into
      a Python dictionary then recursively replace any variable references using the supplied
      env dictionary.
  """

  if config is None:
    return None
  stripped = config.strip()
  if len(stripped) == 0:
    config = {}
  elif stripped[0] == '{':
    config = json.loads(config)
  else:
    config = yaml.load(config)
  if as_dict:
    config = dict(config)

  # Now we need to walk the config dictionary recursively replacing any '$name' vars.
  replace_vars(config, env)
  return config


def validate_config(config, required_keys, optional_keys=None):
  """ Validate a config dictionary to make sure it includes all required keys
      and does not include any unexpected keys.

  Args:
    config: the config to validate.
    required_keys: the names of the keys that the config must have.
    optional_keys: the names of the keys that the config can have.

  Raises:
    Exception if the config is not a dict or invalid.
  """
  if optional_keys is None:
    optional_keys = []
  if not isinstance(config, dict):
    raise Exception('config is not dict type')
  invalid_keys = set(config) - set(required_keys + optional_keys)
  if len(invalid_keys) > 0:
    raise Exception('Invalid config with unexpected keys "%s"'
        % ', '.join(e for e in invalid_keys))
  missing_keys = set(required_keys) - set(config)
  if len(missing_keys) > 0:
    raise Exception('Invalid config with missing keys "%s"' % ', '.join(missing_keys))


def validate_config_must_have(config, required_keys):
  """ Validate a config dictionary to make sure it has all of the specified keys

  Args:
    config: the config to validate.
    required_keys: the list of possible keys that config must include.

  Raises:
    Exception if the config does not have any of them.
  """
  missing_keys = set(required_keys) - set(config)
  if len(missing_keys) > 0:
    raise Exception('Invalid config with missing keys "%s"' % ', '.join(missing_keys))


def validate_config_has_one_of(config, one_of_keys):
  """ Validate a config dictionary to make sure it has one and only one
      key in one_of_keys.

  Args:
    config: the config to validate.
    one_of_keys: the list of possible keys that config can have one and only one.

  Raises:
    Exception if the config does not have any of them, or multiple of them.
  """
  intersection = set(config).intersection(one_of_keys)
  if len(intersection) > 1:
    raise Exception('Only one of the values in "%s" is needed' % ', '.join(intersection))
  if len(intersection) == 0:
    raise Exception('One of the values in "%s" is needed' % ', '.join(one_of_keys))


def validate_config_value(value, possible_values):
  """ Validate a config value to make sure it is one of the possible values.

  Args:
    value: the config value to validate.
    possible_values: the possible values the value can be

  Raises:
    Exception if the value is not one of possible values.
  """
  if value not in possible_values:
    raise Exception('Invalid config value "%s". Possible values are %s'
        % (value, ', '.join(e for e in possible_values)))


# For chart and table HTML viewers, we use a list of table names and reference
# instead the indices in the HTML, so as not to include things like projectID, etc,
# in the HTML.

_data_sources = []


def get_data_source_index(name):
  if name not in _data_sources:
    _data_sources.append(name)
  return _data_sources.index(name)


def validate_gcs_path(path, require_object):
  """ Check whether a given path is a valid GCS path.

  Args:
    path: the config to check.
    require_object: if True, the path has to be an object path but not bucket path.

  Raises:
    Exception if the path is invalid
  """
  bucket, key = datalab.storage._bucket.parse_name(path)
  if bucket is None:
    raise Exception('Invalid GCS path "%s"' % path)
  if require_object and key is None:
    raise Exception('It appears the GCS path "%s" is a bucket path but not an object path' % path)


def parse_control_options(controls, variable_defaults=None):
  """ Parse a set of control options.

  Args:
    controls: The dictionary of control options.
    variable_defaults: If the controls are for a Query with variables, then this is the
        default variable values defined in the Query module. The options in the controls
        parameter can override these but if a variable has no 'value' property then we
        fall back to these.

  Returns:
    - the HTML for the controls.
    - the default values for the controls as a dict.
    - the list of DIV IDs of the controls.

  """
  controls_html = ''
  control_defaults = {}
  control_ids = []
  div_id = _html.Html.next_id()
  if variable_defaults is None:
    variable_defaults = {}
  for varname, control in list(controls.items()):
    label = control.get('label', varname)
    control_id = div_id + '__' + varname
    control_ids.append(control_id)
    value = control.get('value', variable_defaults.get(varname, None))
    # The user should usually specify the type but we will default to 'textbox' for strings
    # and 'set' for lists.
    if isinstance(value, basestring):
      type = 'textbox'
    elif isinstance(value, list):
      type = 'set'
    else:
      type = None
    type = control.get('type', type)

    if type == 'picker':
      choices = control.get('choices', value)
      if not isinstance(choices, list) or len(choices) == 0:
        raise Exception('picker control must specify a nonempty set of choices')
      if value is None:
        value = choices[0]
      choices_html = ''
      for i, choice in enumerate(choices):
        choices_html += "<option value=\"%s\" %s>%s</option>" % \
                        (choice, ("selected=\"selected\"" if choice == value else ''), choice)
      control_html = "{label}<select disabled id=\"{id}\">{choices}</select>" \
          .format(label=label, id=control_id, choices=choices_html)
    elif type == 'set':  # Multi-picker; implemented as checkboxes.
      # TODO(gram): consider using "name" property of the control to group checkboxes. That
      # way we can save the code of constructing and parsing control Ids with sequential
      #  numbers in it. Multiple checkboxes can share the same name.
      choices = control.get('choices', value)
      if not isinstance(choices, list) or len(choices) == 0:
        raise Exception('set control must specify a nonempty set of choices')
      if value is None:
        value = choices
      choices_html = ''
      control_ids[-1] = '%s:%d' % (control_id, len(choices))  # replace ID to include count.
      for i, choice in enumerate(choices):
        checked = choice in value
        choice_id = '%s:%d' % (control_id, i)
        # TODO(gram): we may want a 'Submit/Refresh button as we may not want to rerun
        # query on each checkbox change.
        choices_html += """
          <div>
            <label>
              <input type="checkbox" id="{id}" value="{choice}" {checked} disabled>
              {choice}
            </label>
          </div>
        """.format(id=choice_id, choice=choice, checked="checked" if checked else '')
      control_html = "{label}<div>{choices}</div>".format(label=label, choices=choices_html)
    elif type == 'checkbox':
      control_html = """
            <label>
              <input type="checkbox" id="{id}" {checked} disabled>
              {label}
            </label>
        """.format(label=label, id=control_id, checked="checked" if value else '')
    elif type == 'slider':
      min_ = control.get('min', None)
      max_ = control.get('max', None)
      if min_ is None or max_ is None:
        raise Exception('slider control must specify a min and max value')
      if max_ <= min_:
        raise Exception('slider control must specify a min value less than max value')
      step = control.get('step', 1 if isinstance(min_, int) and isinstance(max_, int)
                         else (float(max_ - min_) / 10.0))
      if value is None:
        value = min_
      control_html = """
        {label}
        <input type="text" class="gchart-slider_value" id="{id}_value" value="{value}" disabled/>
        <input type="range" class="gchart-slider" id="{id}" min="{min}" max="{max}" step="{step}"
            value="{value}" disabled/>
      """.format(label=label, id=control_id, value=value, min=min_, max=max_, step=step)
    elif type == 'textbox':
      if value is None:
        value = ''
      control_html = "{label}<input type=\"text\" value=\"{value}\" id=\"{id}\" disabled/>" \
          .format(label=label, value=value, id=control_id)
    else:
      raise Exception(
          'Unknown control type %s (expected picker, slider, checkbox, textbox or set)' % type)

    control_defaults[varname] = value
    controls_html += "<div class=\"gchart-control\">{control}</div>\n" \
        .format(control=control_html)

  controls_html = "<div class=\"gchart-controls\">{controls}</div>".format(controls=controls_html)
  return controls_html, control_defaults, control_ids


def chart_html(driver_name, chart_type, source, chart_options=None, fields='*', refresh_interval=0,
               refresh_data=None, control_defaults=None, control_ids=None, schema=None):
  """ Return HTML for a chart.

  Args:
    driver_name: the name of the chart driver. Currently we support 'plotly' or 'gcharts'.
    chart_type: string specifying type of chart.
    source: the data source for the chart. Can be actual data (e.g. list) or the name of
        a data source (e.g. the name of a query module).
    chart_options: a dictionary of options for the chart. Can contain a 'controls' entry
        specifying controls. Other entries are passed as JSON to Google Charts.
    fields: the fields to chart. Can be '*' for all fields (only sensible if the columns are
        ordered; e.g. a Query or list of lists, but not a list of dictionaries); otherwise a
        string containing a comma-separated list of field names.
    refresh_interval: a time in seconds after which the chart data will be refreshed. 0 if the
        chart should not be refreshed (i.e. the data is static).
    refresh_data: if the source is a list or other raw data, this is a YAML string containing
        metadata needed to support calls to refresh (get_chart_data).
    control_defaults: the default variable values for controls that are shared across charts
        including this one.
    control_ids: the DIV IDs for controls that are shared across charts including this one.
    schema: an optional schema for the data; if not supplied one will be inferred.

  Returns:
    A string containing the HTML for the chart.

  """
  div_id = _html.Html.next_id()
  controls_html = ''
  if control_defaults is None:
    control_defaults = {}
  if control_ids is None:
    control_ids = []
  if chart_options is not None and 'variables' in chart_options:
    controls = chart_options['variables']
    del chart_options['variables']  # Just to make sure GCharts doesn't see them.
    try:
      item = get_notebook_item(source)
      _, variable_defaults = datalab.data.SqlModule.get_sql_statement_with_environment(item, '')
    except Exception:
      variable_defaults = {}
    controls_html, defaults, ids = parse_control_options(controls, variable_defaults)
    # We augment what we are passed so that in principle we can have controls that are
    # shared by charts as well as controls that are specific to a chart.
    control_defaults.update(defaults)
    control_ids.extend(ids),

  _HTML_TEMPLATE = """
    <div class="bqgc-container">
      {controls}
      <div class="bqgc {extra_class}" id="{id}">
      </div>
    </div>
    <script>

      require.config({{
        paths: {{
          d3: '//cdnjs.cloudflare.com/ajax/libs/d3/3.4.13/d3',
          plotly: 'https://cdn.plot.ly/plotly-1.5.1.min.js?noext',
          jquery: '//ajax.googleapis.com/ajax/libs/jquery/2.0.0/jquery.min'
        }},
        map: {{
          '*': {{
            datalab: 'nbextensions/gcpdatalab'
          }}
        }},
        shim: {{
          plotly: {{
            deps: ['d3', 'jquery'],
            exports: 'plotly'
          }}
        }}
      }});

      require(['datalab/charting',
               'datalab/element!{id}',
               'base/js/events',
               'datalab/style!/nbextensions/gcpdatalab/charting.css'
              ],
        function(charts, dom, events) {{
          charts.render(
              '{driver}',
              dom,
              events,
              '{chart_type}',
              {control_ids},
              {data},
              {options},
              {refresh_data},
              {refresh_interval},
              {total_rows});
          }}
        );
    </script>
  """
  count = 25 if chart_type == 'paged_table' else -1
  data, total_count = get_data(source, fields, control_defaults, 0, count, schema)
  if refresh_data is None:
    if isinstance(source, basestring):
      source_index = get_data_source_index(source)
      refresh_data = {'source_index': source_index, 'name': source_index}
    else:
      refresh_data = {'name': 'raw data'}
  refresh_data['fields'] = fields

  # TODO(gram): check if we need to augment env with user_ns
  return _HTML_TEMPLATE \
      .format(driver=driver_name,
              controls=controls_html,
              id=div_id,
              chart_type=chart_type,
              extra_class=" bqgc-controlled" if len(controls_html) else '',
              data=json.dumps(data, cls=datalab.utils.JSONEncoder),
              options=json.dumps(chart_options, cls=datalab.utils.JSONEncoder),
              refresh_data=json.dumps(refresh_data, cls=datalab.utils.JSONEncoder),
              refresh_interval=refresh_interval,
              control_ids=str(control_ids),
              total_rows=total_count)


def profile_df(df):
  """ Generate a profile of data in a dataframe.

  Args:
    df: the Pandas dataframe.
  """
  # The bootstrap CSS messes up the Datalab display so we tweak it to not have an effect.
  # TODO(gram): strip it out rather than this kludge.
  return IPython.core.display.HTML(
      pandas_profiling.ProfileReport(df).html.replace('bootstrap', 'nonexistent'))



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Context functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

from . import _project
from . import _utils


class Context(object):
  """Maintains contextual state for connecting to Cloud APIs.
  """

  _global_context = None

  def __init__(self, project_id, credentials):
    """Initializes an instance of a Context object.

    Args:
      project_id: the current cloud project.
      credentials: the credentials to use to authorize requests.
    """
    self._project_id = project_id
    self._credentials = credentials

  @property
  def credentials(self):
    """Retrieves the value of the credentials property.

    Returns:
      The current credentials used in authorizing API requests.
    """
    return self._credentials

  def set_credentials(self, credentials):
    """ Set the credentials for the context. """
    self._credentials = credentials

  @property
  def project_id(self):
    """Retrieves the value of the project_id property.

    Returns:
      The current project id to associate with API requests.
    """
    if not self._project_id:
      raise Exception('No project ID found. Perhaps you should set one by running "%projects set <project-id>" in a code cell.')
    return self._project_id

  def set_project_id(self, project_id):
    """ Set the project_id for the context. """
    self._project_id = project_id

  @staticmethod
  def is_signed_in():
    """ If the user has signed in or it is on GCE VM with default credential."""
    try:
      _utils.get_credentials()
      return True
    except Exception:
      return False

  @staticmethod
  def default():
    """Retrieves a default Context object, creating it if necessary.

      The default Context is a global shared instance used every time the default context is
      retrieved.

      Attempting to use a Context with no project_id will raise an exception, so on first use
      set_project_id must be called.

    Returns:
      An initialized and shared instance of a Context object.
    """
    credentials = _utils.get_credentials()
    if Context._global_context is None:
      project = _project.Projects.get_default_id(credentials)
      Context._global_context = Context(project, credentials)
    else:
      # Always update the credentials in case the access token is revoked or expired
      Context._global_context.set_credentials(credentials)
    return Context._global_context



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements BigQuery Job functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import datetime

import datalab.utils

from . import _api


class Job(datalab.utils.GCPJob):
  """Represents a BigQuery Job.
  """

  def __init__(self, job_id, context):
    """Initializes an instance of a Job.

    Args:
      job_id: the BigQuery job ID corresponding to this job.
      context: a Context object providing project_id and credentials.
    """
    super(Job, self).__init__(job_id, context)

  def _create_api(self, context):
    return _api.Api(context)

  def _refresh_state(self):
    """ Get the state of a job. If the job is complete this does nothing;
        otherwise it gets a refreshed copy of the job resource.
    """
    # TODO(gram): should we put a choke on refreshes? E.g. if the last call was less than
    # a second ago should we return the cached value?
    if self._is_complete:
      return

    try:
      response = self._api.jobs_get(self._job_id)
    except Exception as e:
      raise e

    if 'status' in response:
      status = response['status']
      if 'state' in status and status['state'] == 'DONE':
        self._end_time = datetime.datetime.utcnow()
        self._is_complete = True
        self._process_job_status(status)

    if 'statistics' in response:
      statistics = response['statistics']
      start_time = statistics.get('creationTime', None)
      end_time = statistics.get('endTime', None)
      if start_time and end_time and end_time >= start_time:
        self._start_time = datetime.datetime.fromtimestamp(float(start_time) / 1000.0)
        self._end_time = datetime.datetime.fromtimestamp(float(end_time) / 1000.0)

  def _process_job_status(self, status):
    if 'errorResult' in status:
      error_result = status['errorResult']
      location = error_result.get('location', None)
      message = error_result.get('message', None)
      reason = error_result.get('reason', None)
      self._fatal_error = datalab.utils.JobError(location, message, reason)
    if 'errors' in status:
      self._errors = []
      for error in status['errors']:
        location = error.get('location', None)
        message = error.get('message', None)
        reason = error.get('reason', None)
        self._errors.append(datalab.utils.JobError(location, message, reason))



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Support for getting gcloud credentials. """

from __future__ import absolute_import
from __future__ import unicode_literals
import oauth2client.client
import json
import os
import subprocess


# TODO(ojarjur): This limits the APIs against which Datalab can be called
# (when using a service account with a credentials file) to only being those
# that are part of the Google Cloud Platform. We should either extend this
# to all of the API scopes that Google supports, or make it extensible so
# that the user can define for themselves which scopes they want to use.
CREDENTIAL_SCOPES = [
  'https://www.googleapis.com/auth/cloud-platform',
]


def _in_datalab_docker():
  return os.path.exists('/datalab') and os.getenv('DATALAB_ENV')


def get_config_dir():
  config_dir = os.getenv('CLOUDSDK_CONFIG')
  if config_dir is None:
    if os.name == 'nt':
      try:
        config_dir = os.path.join(os.environ['APPDATA'], 'gcloud')
      except KeyError:
        # This should never happen unless someone is really messing with things.
        drive = os.environ.get('SystemDrive', 'C:')
        config_dir = os.path.join(drive, '\\gcloud')
    else:
      config_dir = os.path.join(os.path.expanduser('~'), '.config/gcloud')
  return config_dir


def get_credentials():
  """ Get the credentials to use. We try application credentials first, followed by
      user credentials. The path to the application credentials can be overridden
      by pointing the GOOGLE_APPLICATION_CREDENTIALS environment variable to some file;
      the path to the user credentials can be overridden by pointing the CLOUDSDK_CONFIG
      environment variable to some directory (after which we will look for the file
      $CLOUDSDK_CONFIG/gcloud/credentials). Unless you have specific reasons for
      overriding these the defaults should suffice.
  """
  try:
    credentials = oauth2client.client.GoogleCredentials.get_application_default()
    if credentials.create_scoped_required():
      credentials = credentials.create_scoped(CREDENTIAL_SCOPES)
    return credentials
  except Exception as e:

    # Try load user creds from file
    cred_file = get_config_dir() + '/credentials'
    if os.path.exists(cred_file):
      with open(cred_file) as f:
        creds= json.loads(f.read())
      # Use the first gcloud one we find
      for entry in creds['data']:
        if entry['key']['type'] == 'google-cloud-sdk':
          return oauth2client.client.OAuth2Credentials.from_json(json.dumps(entry['credential']))

    if type(e) == oauth2client.client.ApplicationDefaultCredentialsError:
      # If we are in Datalab container, change the message to be about signing in.
      if _in_datalab_docker():
        raise Exception('No application credentials found. Perhaps you should sign in.')

    raise e


def save_project_id(project_id):
  """ Save project id to config file.

  Args:
    project_id: the project_id to save.
  """
  # Try gcloud first. If gcloud fails (probably because it does not exist), then
  # write to a config file.
  try:
    subprocess.call(['gcloud', 'config', 'set', 'project', project_id])
  except:
    config_file = os.path.join(get_config_dir(), 'config.json')
    config = {}
    if os.path.exists(config_file):
      with open(config_file) as f:
        config = json.loads(f.read())
    config['project_id'] = project_id
    with open(config_file, 'w') as f:
      f.write(json.dumps(config))


def get_project_id():
  """ Get default project id from config or environment var.

  Returns: the project id if available, or None.
  """
  # Try getting default project id from gcloud. If it fails try config.json.
  try:
    proc = subprocess.Popen(['gcloud', 'config', 'list', '--format', 'value(core.project)'],
                            stdout=subprocess.PIPE)
    stdout, _ = proc.communicate()
    value = stdout.strip()
    if proc.poll() == 0 and value:
      return value
  except:
    pass

  config_file = os.path.join(get_config_dir(), 'config.json')
  if os.path.exists(config_file):
    with open(config_file) as f:
      config = json.loads(f.read())
      if 'project_id' in config and config['project_id']:
        return str(config['project_id'])

  if os.getenv('PROJECT_ID') is not None:
    return os.getenv('PROJECT_ID')
  return None



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implementation of various module magics"""
from __future__ import absolute_import
from __future__ import unicode_literals

try:
  import IPython
  import IPython.core.magic
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import sys
import types

from . import _commands
from . import _utils


@IPython.core.magic.register_line_cell_magic
def pymodule(line, cell=None):
  """Creates and subsequently auto-imports a python module.
  """
  parser = _commands.CommandParser.create('pymodule')
  parser.add_argument('-n', '--name',
                      help='the name of the python module to create and import')
  parser.set_defaults(func=_pymodule_cell)
  return _utils.handle_magic_line(line, cell, parser)


def _pymodule_cell(args, cell):
  if cell is None:
      raise Exception('The code for the module must be included')

  name = args['name']
  module = _create_python_module(name, cell)

  # Automatically import the newly created module by assigning it to a variable
  # named the same name as the module name.
  ipy = IPython.get_ipython()
  ipy.push({name: module})


def _create_python_module(name, code):
  # By convention the module is associated with a file name matching the module name
  module = types.ModuleType(str(name))
  module.__file__ = name
  module.__name__ = name

  exec(code, module.__dict__)

  # Hold on to the module if the code executed successfully
  sys.modules[name] = module
  return module



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements job view"""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str


try:
  import IPython
  import IPython.core.magic
  import IPython.core.display
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import datalab.utils

from . import _html


_local_jobs = {}


def html_job_status(job_name, job_type, refresh_interval, html_on_running, html_on_success):
  """create html representation of status of a job (long running operation).

  Args:
    job_name: the full name of the job.
    job_type: type of job. Can be 'local' or 'cloud'.
    refresh_interval: how often should the client refresh status.
    html_on_running: additional html that the job view needs to include on job running.
    html_on_success: additional html that the job view needs to include on job success.
  """
  _HTML_TEMPLATE = """
    <div class="jobstatus" id="%s">
    </div>
    <script>
      require(['datalab/job', 'datalab/element!%s', 'base/js/events',
          'datalab/style!/nbextensions/datalab/job.css'],
        function(job, dom, events) {
          job.render(dom, events, '%s', '%s', %s, '%s', '%s');
        }
      );
    </script>"""
  div_id = _html.Html.next_id()
  return IPython.core.display.HTML(_HTML_TEMPLATE % (div_id, div_id, job_name, job_type,
                                   refresh_interval, html_on_running, html_on_success))


@IPython.core.magic.register_line_magic
def _get_job_status(line):
  """magic used as an endpoint for client to get job status.

       %_get_job_status <name>

  Returns:
    A JSON object of the job status.
  """
  try:
    args = line.strip().split()
    job_name = args[0]

    job = None
    if job_name in _local_jobs:
      job = _local_jobs[job_name]
    else:
      raise Exception('invalid job %s' % job_name)

    if job is not None:
      error = '' if job.fatal_error is None else str(job.fatal_error)
      data = {'exists': True, 'done': job.is_complete, 'error': error}
    else:
      data = {'exists': False}

  except Exception as e:
    datalab.utils.print_exception_with_last_stack(e)
    data = {'done': True, 'error': str(e)}

  return IPython.core.display.JSON(data)



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Cloud ML Summary wrapper."""

import datetime
import glob
import os
from tensorflow.core.util import event_pb2
from tensorflow.python.lib.io import tf_record

import datalab.storage as storage


class Summary(object):
  """Represents TensorFlow summary events from files under a directory."""

  def __init__(self, path):
    """Initializes an instance of a Summary.

    Args:
      path: the path of the directory which holds TensorFlow events files.
            Can be local path or GCS path.
    """
    self._path = path

  def _get_events_files(self):
    if self._path.startswith('gs://'):
      storage._api.Api.verify_permitted_to_read(self._path)
      bucket, prefix = storage._bucket.parse_name(self._path)
      items = storage.Items(bucket, prefix, None)
      filtered_list = [item.uri for item in items if os.path.basename(item.uri).find('tfevents')]
      return filtered_list
    else:
      path_pattern = os.path.join(self._path, '*tfevents*')
      return glob.glob(path_pattern)

  def list_events(self):
    """List all scalar events in the directory.

    Returns:
      A set of unique event tags.
    """
    event_tags = set()
    for event_file in self._get_events_files():
      for record in tf_record.tf_record_iterator(event_file):
        event = event_pb2.Event.FromString(record)
        if event.summary is None or event.summary.value is None:
          continue
        for value in event.summary.value:
          if value.simple_value is None:
            continue
          if value.tag is not None and value.tag not in event_tags:
            event_tags.add(value.tag)
    return event_tags

  def get_events(self, event_name):
    """Get all events of a certain tag.

    Args:
      event_name: the tag of event to look for.

    Returns:
      A tuple. First is a list of {time_span, event_name}. Second is a list of {step, event_name}.

    Raises:
      Exception if event start time cannot be found
    """
    events_time = []
    events_step = []
    event_start_time = None
    for event_file in self._get_events_files():
      for record in tf_record.tf_record_iterator(event_file):
        event = event_pb2.Event.FromString(record)
        if event.file_version is not None:
          # first event in the file.
          time = datetime.datetime.fromtimestamp(event.wall_time)
          if event_start_time is None or event_start_time > time:
            event_start_time = time

        if event.summary is None or event.summary.value is None:
          continue
        for value in event.summary.value:
          if value.simple_value is None or value.tag is None:
            continue
          if value.tag == event_name:
            if event.wall_time is not None:
              time = datetime.datetime.fromtimestamp(event.wall_time)
              events_time.append({'time': time, event_name: value.simple_value})
            if event.step is not None:
              events_step.append({'step': event.step, event_name: value.simple_value})
    if event_start_time is None:
      raise Exception('Empty or invalid TF events file. Cannot find event start time.')
    for event in events_time:
      event['time'] = event['time'] - event_start_time  # convert time to timespan
    events_time = sorted(events_time, key=lambda k: k['time'])
    events_step = sorted(events_step, key=lambda k: k['step'])
    return events_time, events_step



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.


import io
import json
import os
import psutil
import subprocess
import tempfile
import time

import google.cloud.ml as ml

import datalab.utils


def _wait_and_kill(pid_to_wait, pids_to_kill):
  """ Helper function.
      Wait for a process to finish if it exists, and then try to kill a list of processes.
  Args:
    pid_to_wait: the process to wait for.
    pids_to_kill: a list of processes to kill after the process of pid_to_wait finishes.
  """
  if psutil.pid_exists(pid_to_wait):
    psutil.Process(pid=pid_to_wait).wait()

  for pid_to_kill in pids_to_kill:
    if psutil.pid_exists(pid_to_kill):
      p = psutil.Process(pid=pid_to_kill)
      p.kill()
      p.wait()


class LocalRunner(object):
  """Provides a "distributed" local run of a CloudML trainer packaged in a tarball.
     It simulates CloudML service by spawning master, worker, and ps processses,
     but instead of running in their own GKE cluster pods it runs all these as local processes.
  """

  def __init__(self, tar_path, module_to_run, logdir, replica_spec, program_args, all_args):
    """Initializes an instance of a LocalRunner

    Args:
      tar_path: the path of the trainer packaged in a tarball. Can be a local path or a GCS path.
      module_to_run: the module to run in the tarball.
      logdir: the directory to save logs.
      replica_spec: the number of replicas for each job_type.
          For example, {'master': 1, 'worker': 1, 'ps': 1}.
          Currently it supports at most one process for each job_type, and 'master' is required.
      program_args: the arguments of the training job program. For example,
          {
            'train_data_paths': ['/content/mydata/features_train'],
            'eval_data_paths': ['/content/mydata/features_eval'],
            'metadata_path': '/content/mydata/metadata.yaml',
            'output_path': '/content/mymodel/',
          }
      all_args: all args that can be submitted to cloud training such as job name, replicas, etc.
          It is aligned to the CloudML training service interface. In the program, it can be
          retrieved by 'TF_CONFIG' env var (json serialized) under 'job' member.

    Raises:
      Exception if replica_spec does not contain 'master' or its value is below one.
    """
    self._tar_path = tar_path
    self._module_to_run = module_to_run
    self._logdir = logdir
    self._replica_spec = replica_spec
    self._program_args = program_args
    if self._program_args is None:
      self._program_args = {}
    self._all_args = all_args
    if self._all_args is None:
      self._all_args = {}
    self._cluster_spec = self._create_cluster_spec()
    self._task_processes = {}
    self._log_writers = {}
    self._log_readers = {}
    self._monitor_process = None

  def _create_cluster_spec(self):
    """Create cluster spec that will be passed to each task process as command parameter.
       This matches CloudML training service behavior.
    """
    spec = {}
    for job_type, replicas in self._replica_spec.iteritems():
      if replicas > 0:
        port = datalab.utils.pick_unused_port()
        spec[job_type] = ['localhost:' + str(port)]
    if 'master' not in spec:
      raise Exception('Need to have at least 1 master replica')
    return spec

  def _create_tf_config(self, job_type):
    """Create a list of arguments that will be passed to task process as command
       parameters. This matches CloudML training service behavior.
    """
    task_spec = {'type': job_type, 'index': 0}
    return {
      'cluster': self._cluster_spec,
      'task': task_spec,
      'job': self._all_args,
    }

  def _create_task_args(self):
    args = [
      'python',
      '-m',
      self._module_to_run,
    ]
    for k,v in self._program_args.iteritems():
      if isinstance(v, list):
        for item in v:
          args.append('--' + k)
          args.append(str(item))
      else:
        args.append('--' + k)
        args.append(str(v))
    return args

  def _extract_tar(self):
    extracted_dir = tempfile.mkdtemp()
    tar_path = self._tar_path
    if tar_path.startswith('gs://'):
      tar_path = os.path.join(extracted_dir, os.path.basename(tar_path))
      ml.util._file.copy_file(self._tar_path, tar_path)
    subprocess.check_call(['pip', 'install', tar_path, '--target', extracted_dir,
                           '--upgrade', '--force-reinstall'])
    return extracted_dir

  def _clean_up(self):
    processes_to_clean = list(self._task_processes.values())
    if self._monitor_process is not None:
      processes_to_clean.append(self._monitor_process)

    for p in processes_to_clean:
      if p.poll() is None:
        # TODO(qimingj): consider p.kill() if it does not terminate in certain time.
        p.terminate()
        p.wait()

    for k,v in self._log_readers.iteritems():
      v.close()
    for k,v in self._log_writers.iteritems():
      v.close()

  def _start_externel_monitor_process(self):
    """Create a process that monitors the current process. If the current process exists,
       Clean up a list of target processes.
       This is needed to kill all running training processes when the kernel is restarted.
       Note that Jupyter does not kill child processes when kernel is restarted. "atexit"
       hook doesn't work either if the kernel is busy such as in time.sleep (seems SIGKILL
       is used to restart the kernel).
    """
    pids_to_kill = [p.pid for p in self._task_processes.values()]
    script = 'import %s; %s._wait_and_kill(%s, %s)' % \
        (__name__, __name__, str(os.getpid()), str(pids_to_kill))
    self._monitor_process = subprocess.Popen(['python', '-c', script])

  def _start_task_process(self, workdir, job_type):
    args = self._create_task_args()
    logfile = os.path.join(self._logdir, job_type)
    # We cannot pipe child process's stdout and stderr directly because
    # we need to append 'master', 'worker', 'ps' at begining of each
    # log entry. A memory stream such as StringIO does not work here because
    # Popen expects file descriptor. Therefore physical files are used to back
    # up the messages.
    w = io.open(logfile, 'w')
    r = io.open(logfile, 'r')
    tf_config = self._create_tf_config(job_type)
    env = os.environ.copy()
    env['TF_CONFIG'] = json.dumps(tf_config)
    p = subprocess.Popen(args, env=env, stdout=w, stderr=w)
    self._log_writers[job_type] = w
    self._log_readers[job_type] = r
    self._task_processes[job_type] = p

  def _print_task_output(self, callback, param, done):
    if callback is None:
      return
    new_msgs = []
    for job_type, reader in self._log_readers.iteritems():
      content = reader.read()
      if (content):
        lines = content.split('\n')
        for line in lines:
          new_msgs.append(job_type + ': '+ line)
    callback(self._replica_spec, new_msgs, done, param)

  def run(self, callback, param, interval):
    """Run a training job locally. Block the caller until it finishes.
       Prints out task processes stdout and stderr.

    Args:
      callback: a callback that will be invoked every "interval" seconds. The signature is:
          callback(replica_spec, new_msgs, done, param), where:
              replica_spec: the replica spec of the runner
              new_msgs: new output messages that are available from all task processes
              done: whether the job is finished
              param: a callback param.
      param: the callback param that will be passed along whenever the callback is invoked.
      interval: the interval in seconds controlling the callback frequency.
    """
    workdir = self._extract_tar()
    previous_cwd = os.getcwd()
    os.chdir(workdir)
    for job_type, replicas in self._replica_spec.iteritems():
      if replicas > 0:
        self._start_task_process(workdir, job_type)
    os.chdir(previous_cwd)
    self._start_externel_monitor_process()
    while self._task_processes['master'].poll() is None:
      self._print_task_output(callback, param, False)
      time.sleep(interval)
    self._print_task_output(callback, param, True)
    self._clean_up()



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

try:
  import IPython
  import IPython.core.magic
except ImportError:
  raise Exception('This module can only be loaded in ipython.')


import collections
import fnmatch
import google.cloud.ml
import json
import math
import os
import plotly.graph_objs as go
from plotly.offline import iplot
import urllib
import yaml

import datalab.context
import datalab.data
import datalab.mlalpha
import datalab.utils.commands


@IPython.core.magic.register_line_cell_magic
def mlalpha(line, cell=None):
  """Implements the ml line cell magic.

  Args:
    line: the contents of the ml line.
    cell: the contents of the ml cell.

  Returns:
    The results of executing the cell.
  """
  parser = datalab.utils.commands.CommandParser(prog="mlalpha", description="""
Execute various ml-related operations. Use "%%mlalpha <command> -h" for help on a specific command.
""")
  train_parser = parser.subcommand('train', 'Run a training job.')
  train_parser.add_argument('--cloud',
                            help='Whether to run the training job in the cloud.',
                            action='store_true', default=False)
  train_parser.set_defaults(func=_train)
  jobs_parser = parser.subcommand('jobs', 'List jobs in a project.')
  jobs_parser.add_argument('--count',
                           help='The number of jobs to browse from head, default to 10.')
  jobs_parser.add_argument('--filter', help='Filter on jobs.')
  jobs_parser.add_argument('--name',
                           help='The name of the operation to retrieve. If provided, show ' +
                                'detailed information of the operation')
  jobs_parser.add_argument('--trials',
                            help='Whether to show hyperparams tuning graph.',
                            action='store_true', default=False)
  jobs_parser.set_defaults(func=_jobs)
  summary_parser = parser.subcommand('summary', 'List or view summary events.')
  summary_parser.add_argument('--dir',
                              help='A list of dirs to look for events. Can be local or GCS path.',
                              nargs='+', required=True)
  summary_parser.add_argument('--name',
                              help='Names of the summary event. If provided, ' +
                                   'plot specified events in same graph (so make sure their ' +
                                   'units match). Otherwise, list all the unique event ' +
                                   'names from files in the directory.', nargs='*')
  summary_parser.add_argument('--time', help='Whether to plot time events only.',
                              action='store_true', default=False)
  summary_parser.add_argument('--step', help='Whether to plot step events only.',
                              action='store_true', default=False)
  summary_parser.set_defaults(func=_summary)
  features_parser = parser.subcommand('features', 'Generate featureset class template.')
  features_parser.set_defaults(func=_features)
  predict_parser = parser.subcommand('predict', 'Get prediction results given data instances.')
  predict_parser.add_argument('--cloud',
                              help='Whether to run the prediction in the cloud.',
                              action='store_true', default=False)
  predict_parser.add_argument('--model',
                              help='Model identifier. In local prediction, it is the path to ' +
                                   'a model directory. In cloud prediction (--cloud), it is ' +
                                   'model.version.', required=True)
  predict_parser.add_argument('--label',
                              help='In classification scenario, which output in the graph ' +
                                   'is the label index. If provided, the index will be ' +
                                   'converted to label string.')
  predict_parser.add_argument('--data',
                              help='The instance data used to predict. It can be a dataframe ' +
                                   'or a list defined in another cell. If not provided, the ' +
                                   'data needs to be provided in current cell input.')
  predict_parser.add_argument('--project',
                              help='The project for the cloud model to use. if not provided, ' +
                                   'current project is used. Only needed in cloud prediction ' +
                                   '(--cloud)')
  predict_parser.set_defaults(func=_predict)
  model_parser = parser.subcommand('model', 'List or view models.')
  model_parser.add_argument('--name',
                            help='The name and version of the model. If "model.version", ' +
                                 'display the details of the model version. If "model", ' +
                                 'list the versions of the model. If not provided, list ' +
                                 'models under the project.')
  model_parser.add_argument('--project',
                            help='The project under which it looks for models. if not ' +
                                 'provided, current project is used.')
  model_parser.set_defaults(func=_model)
  deploy_parser = parser.subcommand('deploy', 'Deploy a model.')
  deploy_parser.add_argument('--name',
                             help='The name and version of the model in the form of ' +
                                  '"model.version".', required=True)
  deploy_parser.add_argument('--path',
                             help='The Google Cloud Storage path of the directory that ' +
                                  'contains an exported model.', required=True)
  deploy_parser.add_argument('--project',
                             help='The project under which the model will be deployed. if not ' +
                                  'provided, current project is used.')
  deploy_parser.set_defaults(func=_deploy)
  delete_parser = parser.subcommand('delete', 'Delete a model or a model version.')
  delete_parser.add_argument('--name',
                             help='The name and version of the model. If "model.version", ' +
                                  'delete the model version. If "model", delete the model.',
                             required=True)
  delete_parser.add_argument('--project',
                             help='The project under which the model or version will be ' +
                                  'deleted. if not provided, current project is used.')
  delete_parser.set_defaults(func=_delete)
  preprocess_parser = parser.subcommand('preprocess', 'Generate preprocess code template.')
  preprocess_parser.add_argument('--cloud',
                                 help='Whether to produce code running in cloud.',
                                 action='store_true', default=False)
  preprocess_parser.set_defaults(func=_preprocess)
  evaluate_parser = parser.subcommand('evaluate', 'Generate evaluate code template.')
  evaluate_parser.add_argument('--cloud',
                               help='Whether to produce code running in cloud.',
                               action='store_true', default=False)
  evaluate_parser.set_defaults(func=_evaluate)
  dataset_parser = parser.subcommand('dataset', 'Define dataset to explore data.')
  dataset_parser.add_argument('--name',
                              help='The name of the dataset to define.', required=True)
  dataset_parser.set_defaults(func=_dataset)
  module_parser = parser.subcommand('module', 'Define a trainer module.')
  module_parser.add_argument('--name', help='The name of the module.', required=True)
  module_parser.add_argument('--main',
                             help='Whether this module is has main function in the trainer ' +
                                  'package.',
                             action='store_true', default=False)
  module_parser.set_defaults(func=_module)
  package_parser = parser.subcommand('package','Create a trainer package from all modules ' +
                                     'defined with %%mlalpha module.')
  package_parser.add_argument('--name', help='The name of the package.', required=True)
  package_parser.add_argument('--output', help='the output dir of the package.', required=True)
  package_parser.set_defaults(func=_package)
  namespace = datalab.utils.commands.notebook_environment()
  return datalab.utils.commands.handle_magic_line(line, cell, parser, namespace=namespace)


def _get_replica_count(config):
  worker_count = config.get('worker_count', 0)
  parameter_server_count = config.get('parameter_server_count', 0)
  if worker_count > 0 or parameter_server_count > 0:
    return 1, worker_count, parameter_server_count
  scale_tier = config.get('scale_tier', 'BASIC')
  if scale_tier == 'BASIC':
    return 1, 0, 0
  else:
    return 1, 1, 1


def _local_train_callback(replica_spec, new_msgs, done, all_msgs):
  if new_msgs:
    all_msgs += new_msgs
    del all_msgs[0:-20]
    IPython.display.clear_output()
    IPython.display.display_html('<p>Job Running...</p>', raw=True)
    log_file_html = ''
    log_url_prefix = ''
    if datalab.context._utils._in_datalab_docker():
      log_url_prefix = '/_nocachecontent/'
    for job_type, replicas in replica_spec.iteritems():
      if replicas > 0:
        log_file_html += ('<a href="%s" target="_blank">%s log</a>&nbsp;&nbsp;'
                          % (log_url_prefix + job_type, job_type))
    IPython.display.display_html(log_file_html, raw=True)
    IPython.display.display_html('<br/>'.join(all_msgs), raw=True)
  if done:
    IPython.display.display_html('<p>Job Finished.</p>', raw=True)


def _output_train_template():
  content = """%%mlalpha train [--cloud]
package_uris: gs://your-bucket/my-training-package.tar.gz
python_module: your_program.your_module
scale_tier: BASIC
region: us-central1
args:
  string_arg: value
  int_arg: value
  appendable_arg:
    - value1
    - value2
"""
  IPython.get_ipython().set_next_input(content)
  parameters = ['package_uris', 'python_module', 'scale_tier', 'region', 'args']
  required_local = [False, False, False, False, False]
  required_cloud = [True, True, True, True, False]
  description = [
    'A GCS or local (for local run only) path to your python training program package.',
    'The module to run.',
    'Type of resources requested for the job. On local run, BASIC means 1 master process only, ' +
      'and any other values mean 1 master 1 worker and 1 ps processes. But you can also ' +
      'override the values by setting worker_count and parameter_server_count. ' +
      'On cloud, see service definition for possible values.',
    'Where the training job runs. For cloud run only.',
    'Args that will be passed to your training program.'
  ]
  data = [{'Parameters': x[0], 'Local Run Required': str(x[1]),
           'Cloud Run Required': str(x[2]), 'Description': x[3]}
          for x in zip(parameters, required_local, required_cloud, description)]
  html = ('<p>A training input template is created in next cell for you. See cell input ' +
          'instructions below.</p>')
  html += datalab.utils.commands.HtmlBuilder.render_table(data,
      ['Parameters', 'Local Run Required', 'Cloud Run Required', 'Description'])

  return IPython.core.display.HTML(html)


def _train(args, cell):
  """ Train a model. """
  if not cell:
    return _output_train_template()

  env = datalab.utils.commands.notebook_environment()
  config = datalab.utils.commands.parse_config(cell, env)
  if args['cloud']:
    datalab.utils.commands.validate_config_must_have(config,
        ['package_uris', 'python_module', 'scale_tier', 'region'])
    runner = datalab.mlalpha.CloudRunner(config)
    job_info = runner.run()
    job_short_name = job_info['jobId']
    html = '<p>Job "%s" was submitted successfully.<br/>' % job_short_name
    html += 'Run "%%mlalpha jobs --name %s" to view the status of the job.</p>' % job_short_name
    log_url_query_strings = {
      'project': datalab.context.Context.default().project_id,
      'resource': 'ml.googleapis.com/job_id/' + job_short_name
    }
    log_url = 'https://console.developers.google.com/logs/viewer?' + \
        urllib.urlencode(log_url_query_strings)
    html += '<p>Click <a href="%s" target="_blank">here</a> to view cloud log. <br/>' % log_url
    html += 'Start TensorBoard by running "%tensorboard start --logdir=&lt;YourLogDir&gt;".</p>'
    return IPython.core.display.HTML(html)
  else:
    # local training
    package_path = None
    if 'package_uris' not in config:
      if '_ml_modules_' not in env:
        raise Exception('Expect either modules defined with "%%mlalpha module", ' +
                        'or "package_uris" in cell.')
      if '_ml_modules_main_' not in env:
        raise Exception('Expect one ml module defined with "--main flag" as the python ' +
                        'program entry point.')
      package_path = datalab.mlalpha.Packager().package(env['_ml_modules_'], 'trainer')
      config['package_uris'] = package_path
      config['python_module'] = 'trainer.' + env['_ml_modules_main_']

    trainer_uri = config['package_uris']
    module_name = config['python_module']
    masters, workers, parameter_servers = _get_replica_count(config)
    replica_spec = {'master': masters, 'worker': workers, 'ps': parameter_servers}
    all_messages = []
    log_dir = os.getcwd()
    if datalab.context._utils._in_datalab_docker():
      log_dir = '/datalab/nocachecontent'
      if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    program_args = config.get('args', None)
    runner = datalab.mlalpha.LocalRunner(trainer_uri, module_name, log_dir, replica_spec, program_args, config)
    runner.run(_local_train_callback, all_messages, 3)
    if package_path is not None:
      os.remove(package_path)


def _plot_hyperparams_tuning(training_input, training_output):
  if ('hyperparameters' not in training_input or 'trials' not in training_output or
      len(training_output['trials']) == 0):
    print 'No trials found. Maybe none of the trials has completed'
    return

  maximize = training_input['hyperparameters']['goal'] == 'MAXIMIZE'
  hyperparam_scales = {}
  for param in training_input['hyperparameters']['params']:
    hyperparam_scales[param['parameterName']] = param.get('scaleType', '')
  instances = []
  for trial in training_output['trials']:
    if 'finalMetric' not in trial:
      continue
    instance = collections.OrderedDict()
    instance.update({'Objective': trial['finalMetric']['objectiveValue']})
    instance.update({'Trial': trial['trialId']})
    instance.update({'Training Step': trial['finalMetric']['trainingStep']})
    hyperparams = dict(trial['hyperparameters'])
    for k in trial['hyperparameters'].keys():
      if hyperparam_scales.get(k, '') == 'UNIT_LOG_SCALE':
        hyperparams[k + '(log)'] = math.log10(float(hyperparams[k]))
    instance.update(hyperparams)
    instances.append(instance)
  if len(instances) == 0:
    print 'No finalMetric found in any trials. '
    return

  instances_sorted = sorted(instances, key=lambda k: k['Objective'], reverse=maximize)
  # Convert list of dictionary to dictionary of list so it is more compact.
  data = instances_sorted[0]
  for k in data.keys():
    data[k] = [d[k] for d in instances_sorted]

  HTML_TEMPLATE = """
<div id="%s" class="parcoords" style="height:300px;"></div>
<div id="%s" class="parcoords_grid"></div>
<script>
  require.config({
    paths: {
      d3: 'http://d3js.org/d3.v3.min',
      sylvester: '/nbextensions/gcpdatalab/extern/sylvester',
      parcoords: '/nbextensions/gcpdatalab/extern/d3.parcoords',
    },
    shim: {
      parcoords: {
        deps: ['d3', 'sylvester'],
        exports: 'd3'
      },
    }
  });
  require(['parcoords',
           '/nbextensions/gcpdatalab/parcoords.js',
           'nbextensions/gcpdatalab/style!/nbextensions/gcpdatalab/extern/d3.parcoords.css'],
          function(d3, lib) {
            var data = %s;
            lib.plot(d3, %s, %s, data, '%s', '%s');
          });
</script>
"""
  graph_id = 'v' + datalab.utils.commands.Html.next_id()
  grid_id = 'g' + datalab.utils.commands.Html.next_id()
  color_range_string = json.dumps([min(data['Objective']), max(data['Objective'])])
  data_string = json.dumps(data)
  maximize_string = json.dumps(maximize)
  html = HTML_TEMPLATE % (graph_id, grid_id, data_string, color_range_string, maximize_string, graph_id, grid_id)
  return IPython.core.display.HTML(html)


def _jobs(args, _):
  """ List the ML jobs in a project. """
  jobs = datalab.mlalpha.Jobs(filter=args['filter'])
  if args['name']:
    job = jobs.get_job_by_name(args['name'])
    if args['trials']:
      if ('trainingInput' not in job.info or 'trainingOutput' not in job.info):
        print 'job %s doesn\'t seem like a hyperparameter tuning training job.' % args['name']
        return
      return _plot_hyperparams_tuning(job.info['trainingInput'], job.info['trainingOutput'])
    else:
      job_yaml = yaml.safe_dump(job.info)
    return datalab.utils.commands.render_text(job_yaml, preformatted=True)
  else:
    count = int(args['count'] or 10)
    data = []
    for job in jobs:
      if count <= 0:
        break
      count -= 1
      data.append({'Id': job['jobId'], 'State': job.get('state', 'UNKNOWN')})
    return datalab.utils.commands.render_dictionary(data, ['Id', 'State'])


def _plot(data, x_name, x_title, y_names):
  y_title = ','.join(y_names)
  layout = go.Layout(
    title=y_title,
    xaxis=dict(
      title=x_title,
    ),
    yaxis=dict(
      title=y_title,
    )
  )
  plot_data = []
  for trace_name, trace_data in data.iteritems():
    for y_name, events in zip(y_names, trace_data):
      x = [d[x_name] for d in events]
      y = [d[y_name] for d in events]
      plot_data.append({'x': x, 'y': y, 'name': y_name + '-' + trace_name})
  fig = go.Figure(data=plot_data, layout=layout)
  iplot(fig)


def get_dirs(pattern):
  """ Finds all matching dirs given a dir path
      for example:
        input: gs://mybucket/iris/hp/*/summaries
        output: [gs://mybucket/iris/hp/1/summaries, gs://mybucket/iris/hp/2/summaries]
  Args:
    pattern: string pattern for the directory path

  Returns:
    List of all matching directories
  """
  dirs = set()
  path = pattern.rstrip('/')
  for p in google.cloud.ml.util._file.glob_files(path + '/*'):
    dir = None
    while True:
      p = p[:p.rfind('/')]
      if fnmatch.fnmatch(p, path):
        dir = p
      else:
        break
    if dir:
      dirs.add(dir)
  return list(dirs)


def _summary(args, _):
  """ Display summary events in a directory. """
  dirs = args['dir']
  event_names = args['name']
  if event_names is not None and len(event_names) > 0:
    time_data = {}
    step_data = {}
    dir_index = 0
    for dir_pattern in dirs:
      for dir in get_dirs(dir_pattern):
        dir_index += 1
        summary = datalab.mlalpha.Summary(dir)
        trace_events_time = []
        trace_events_step = []
        for event_name in event_names:
          events_time, events_step = summary.get_events(event_name)
          for e in events_time:
            e['time'] = e['time'].total_seconds()
          trace_events_time.append(events_time)
          trace_events_step.append(events_step)
        # Try to find 'label' file under the dir. If found, use the content as label.
        # Otherwise, use dir name as label.
        label = dir
        label_file = os.path.join(dir, 'label')
        if google.cloud.ml.util._file.file_exists(label_file) == True:
          label = 'dir%d/' % dir_index + google.cloud.ml.util._file.load_file(label_file)
        time_data[label] = trace_events_time
        step_data[label] = trace_events_step
    if (not args['time'] and not args['step']) or args['time']:
      _plot(time_data, 'time', 'seconds', event_names)
    if (not args['time'] and not args['step']) or args['step']:
      _plot(step_data, 'step', 'step', event_names)
  else:
    event_names = []
    for dir_pattern in dirs:
      for dir in get_dirs(dir_pattern):
        summary = datalab.mlalpha.Summary(dir)
        event_names += summary.list_events()
    event_names = list(set(event_names))  # remove duplicates
    return datalab.utils.commands.render_list(event_names)


def _output_features_template():
  content = """%%mlalpha features
path: REQUIRED_Fill_In_Gcs_or_Local_Path
headers: List_Of_Column_Names_Seperated_By_Comma
target: REQUIRED_Fill_In_Name_Or_Index_Of_Target_Column
id: Fill_In_Name_Or_Index_Of_Id_Column
format: csv_or_tsv
"""
  IPython.get_ipython().set_next_input(content, replace=True)


def _features(args, cell):
  """ Generate FeatureSet Class From Data"""
  if not cell:
    _output_features_template()
    return

  env = datalab.utils.commands.notebook_environment()
  config = datalab.utils.commands.parse_config(cell, env)
  datalab.utils.commands.validate_config(config, ['path', 'target'],
      optional_keys=['headers', 'id', 'format'])
  format = config.get('format', 'csv')
  # For now, support CSV and TSV only.
  datalab.utils.commands.validate_config_value(format, ['csv', 'tsv'])
  delimiter = ',' if format == 'csv' else '\t'
  csv = datalab.data.Csv(config['path'], delimiter=delimiter)
  headers = None
  if 'headers' in config:
    headers = [e.strip() for e in config['headers'].split(',')]
  df = csv.browse(max_lines=100, headers=headers)
  command = '%%mlalpha features\n' + cell
  _output_featureset_template(df.dtypes, config['target'], config.get('id', None), command)


def _output_featureset_template(dtypes, target_column, id_column, command):
  if target_column not in dtypes:
    if type(target_column) is int:
      target_column = dtypes.keys()[target_column]
    else:
      raise Exception('Column "%s" not found. It can be a name in headers, or an index number.'
                      % target_column)
  if id_column is not None and id_column not in dtypes:
    if type(id_column) is int:
      id_column = dtypes.keys()[id_column]
    else:
      raise Exception('Column "%s" not found. It can be a name in headers, or an index number.'
                      % id_column)
  is_regression = str(dtypes[target_column]).startswith('int') or \
      str(dtypes[target_column]).startswith('float')
  scenario = 'continuous' if is_regression == True else 'discrete'
  columns_remaining = dict(dtypes)
  command_lines = command.split('\n')
  # add spaces in the beginning so they are aligned with others.
  command_formatted = '\n'.join(['        ' + line for line in command_lines])
  content = """import google.cloud.ml.features as features

class CsvFeatures(object):
  \"\"\"This class is generated from command line:
%s
     Please modify it as appropriate!!!
  \"\"\"
  csv_columns = (%s)
  %s = features.target('%s').%s()
""" % (command_formatted, ','.join(["'" + e + "'" for e in dtypes.keys()]),
       target_column.replace('-', '_'), target_column, scenario)
  del columns_remaining[target_column]

  if id_column is not None:
    content += """  %s = features.key('%s')\n""" % (id_column.replace('-', '_'), id_column)
    del columns_remaining[id_column]

  text_columns = [k for k,v in columns_remaining.iteritems() if str(v) == 'object']
  categorical_columns = [k for k,v in columns_remaining.iteritems() if str(v) == 'category']
  numeric_columns = [k for k in columns_remaining.keys() if k not in text_columns and
                     k not in categorical_columns]
  if len(numeric_columns) + len(categorical_columns) > 0:
    content += """  attrs = [\n"""
    for numeric_name in numeric_columns:
      content += """      features.numeric('%s').identity(),\n""" % numeric_name
    for categorical_name in categorical_columns:
      content += """      features.categorical('%s'),\n""" % categorical_name
    content += """  ]\n"""
  if len(text_columns) > 0:
    for text_name in text_columns:
      content += """  %s = features.text('%s').bag_of_words(vocab_size=10000)\n\n""" % \
          (text_name.replace('-', '_'), text_name)
  IPython.get_ipython().set_next_input(content, replace=True)


def _predict(args, cell):
  if args['data'] is not None:
    instances = datalab.utils.commands.get_notebook_item(args['data'])
    if instances is None:
      raise Exception('Data "%s" is not defined' % args['data'])
  elif cell is not None:
    instances = []
    lines = cell.split('\n')
    for line in lines:
      instances.append(line)
  else:
    raise Exception('Expect instance data. Can be provided in input cell, or through '
                    '--data args.')
  if args['cloud']:
    parts = args['model'].split('.')
    if len(parts) != 2:
      raise Exception('Invalid model name for cloud prediction. Use "model.version".')
    lp = datalab.mlalpha.CloudPredictor(parts[0], parts[1],
                                   label_output=args['label'],
                                   project_id=args['project'])
  else:
    lp = datalab.mlalpha.LocalPredictor(args['model'],
                                   label_output=args['label'])
  return datalab.utils.commands.render_text(yaml.safe_dump(lp.predict(instances),
                                                           default_flow_style=False),
                                            preformatted=True)


def _model(args, _):
  if args['name'] is None:
    data = list(datalab.mlalpha.CloudModels(project_id=args['project']))
    if len(data) > 0:
      return datalab.utils.commands.render_dictionary(data, data[0].keys())
    print 'No models found.'
    return

  parts = args['name'].split('.')
  if len(parts) == 1:
    data = list(datalab.mlalpha.CloudModelVersions(parts[0], project_id=args['project']))
    if len(data) > 0:
      return datalab.utils.commands.render_dictionary(data, data[0].keys())
    print 'No versions found.'
    return
  elif len(parts) == 2:
    versions = datalab.mlalpha.CloudModelVersions(parts[0], project_id=args['project'])
    version_yaml = yaml.safe_dump(versions.get(parts[1]))
    return datalab.utils.commands.render_text(version_yaml, preformatted=True)
  else:
    raise Exception('Too many "." in name. Use "model" or "model.version".')


def _deploy(args, _):
  parts = args['name'].split('.')
  if len(parts) != 2:
    raise Exception('Invalid model name. Use "model.version".')
  versions = datalab.mlalpha.CloudModelVersions(parts[0], project_id=args['project'])
  versions.deploy(parts[1], args['path'])


def _delete(args, _):
  parts = args['name'].split('.')
  if len(parts) == 1:
    models = datalab.mlalpha.CloudModels(project_id=args['project'])
    models.delete(parts[0])
  elif len(parts) == 2:
    versions = datalab.mlalpha.CloudModelVersions(parts[0], project_id=args['project'])
    versions.delete(parts[1])
  else:
    raise Exception('Too many "." in name. Use "model" or "model.version".')


def _output_preprocess_template(is_cloud):
  content = """%%mlalpha preprocess"""
  if is_cloud:
    content += ' --cloud'
  content += """
train_data_path: REQUIRED_Fill_In_Training_Data_Path
eval_data_path: Fill_In_Evaluation_Data_Path
data_format: REQUIRED_CSV_or_JSON
output_dir: REQUIRED_Fill_In_Output_Path
feature_set_class_name: REQUIRED_Fill_In_FeatureSet_Class_name
"""
  IPython.get_ipython().set_next_input(content, replace=True)


def _pipeline_definition_code(is_cloud, job_name_prefix):
  # TODO: Remove 'extra_packages' once it is not needed by dataflow.
  if is_cloud:
    content_pipeline = \
"""import datetime
options = {
    'staging_location': os.path.join(OUTPUT_DIR, 'tmp', 'staging'),
    'temp_location': os.path.join(OUTPUT_DIR, 'tmp'),
    'job_name': '%s' + '-' + datetime.datetime.now().strftime('%%y%%m%%d-%%H%%M%%S'),
    'project': '%s',
    'extra_packages': [ml.sdk_location],
    'teardown_policy': 'TEARDOWN_ALWAYS',
    'no_save_main_session': True
}
opts = beam.pipeline.PipelineOptions(flags=[], **options)
pipeline = beam.Pipeline('DataflowPipelineRunner', options=opts)
""" % (job_name_prefix, datalab.context.Context.default().project_id)
  else:
    content_pipeline = """pipeline = beam.Pipeline('DirectPipelineRunner')\n"""

  return content_pipeline


def _header_code(command):
  header = \
"""\"\"\"
Following code is generated from command line:
%s\n
Please modify as appropriate!!!
\"\"\"
""" % command
  return header


def _output_preprocess_code_template(command, is_cloud, data_format, train_data_path,
                                     output_dir, feature_set_class_name, eval_data_path=None):
  content_header = _header_code(command)

  content_imports = \
"""import apache_beam as beam
import google.cloud.ml as ml
import google.cloud.ml.io as io
import os
"""

  if data_format == 'CSV':
    coder = """io.CsvCoder.from_feature_set(feature_set, feature_set.csv_columns)"""
  else: # JSON
    coder = """io.JsonCoder.from_feature_set(feature_set)"""
  job_name_prefix = 'preprocess-' + feature_set_class_name.lower().replace('_', '-')

  content_defines = \
"""feature_set = %s()
OUTPUT_DIR = '%s'
%s
""" % (feature_set_class_name, output_dir, _pipeline_definition_code(is_cloud, job_name_prefix))

  content_preprocessing = \
"""training_data = beam.io.TextFileSource(
    '%s',
    strip_trailing_newlines=True,
    coder=%s)
train = pipeline | beam.Read('ReadTrainingData', training_data)
""" % (train_data_path, coder)

  if eval_data_path is not None:
    content_preprocessing += \
"""
eval_data = beam.io.TextFileSource(
    '%s',
    strip_trailing_newlines=True,
    coder=%s)
eval = pipeline  | beam.Read('ReadEvalData', eval_data)
""" % (eval_data_path, coder)

  eval_features = ', eval_features' if eval_data_path is not None else ''
  eval_input = ', eval' if eval_data_path is not None else ''

  content_preprocessing += \
"""
(metadata, train_features%s) = ((train%s) | 'Preprocess'
    >> ml.Preprocess(feature_set, input_format='csv',
                  format_metadata={'headers': feature_set.csv_columns}))
""" % (eval_features, eval_input)

  content_preprocessing += \
"""
(metadata        | 'SaveMetadata'
    >> io.SaveMetadata(os.path.join(OUTPUT_DIR, 'metadata.yaml')))
"""

  content_preprocessing += \
"""
(train_features  | 'SaveTrain'
    >> io.SaveFeatures(os.path.join(OUTPUT_DIR, 'features_train'), shard_name_template=''))
"""

  if eval_data_path is not None:
    content_preprocessing += \
"""
(eval_features   | 'SaveEval'
    >> io.SaveFeatures(os.path.join(OUTPUT_DIR, 'features_eval'), shard_name_template=''))
"""

  content_run = """pipeline.run()"""

  content = \
"""
# header
%s
# imports
%s
# defines
%s
# preprocessing
%s
# run pipeline
%s
""" % (content_header, content_imports, content_defines, content_preprocessing, content_run)
  IPython.get_ipython().set_next_input(content, replace=True)


def _preprocess(args, cell):
  if not cell:
    _output_preprocess_template(args['cloud'])
    return

  env = datalab.utils.commands.notebook_environment()
  config = datalab.utils.commands.parse_config(cell, env)
  datalab.utils.commands.validate_config(config,
     ['train_data_path', 'data_format', 'output_dir', 'feature_set_class_name'],
     optional_keys=['eval_data_path'])
  datalab.utils.commands.validate_config_value(config['data_format'], ['CSV', 'JSON'])
  command = '%%mlalpha preprocess'
  if args['cloud']:
    command += ' --cloud'
  command += '\n' + cell
  _output_preprocess_code_template(command, args['cloud'], config['data_format'],
      config['train_data_path'], config['output_dir'], config['feature_set_class_name'],
      eval_data_path=config.get('eval_data_path', None))


def _output_evaluate_template(is_cloud):
  content = """%%mlalpha evaluate"""
  if is_cloud:
    content += ' --cloud'
  content += """
preprocessed_eval_data_path: REQUIRED_Fill_In_Eval_Data_Path
metadata_path: REQUIRED_Fill_In_Metadata_Path
model_dir: REQUIRED_Fill_In_Model_Path
output_dir: REQUIRED_Fill_In_Output_Path
output_prediction_name: Fill_In_Prediction_Output
output_score_name: Fill_In_Score_Output
"""
  IPython.get_ipython().set_next_input(content, replace=True)


def _output_evaluate_code_template(command, is_cloud, preprocessed_eval_data_path,
                                   metadata_path, model_dir, output_dir,
                                   output_prediction_name=None, output_score_name=None):
  # output_prediction_name is only useful for generating results analysis code.
  # It is only used in classification but not regression.
  content_header = _header_code(command)

  content_imports = \
"""import apache_beam as beam
import google.cloud.ml as ml
import google.cloud.ml.analysis as analysis
import google.cloud.ml.io as io
import json
import os
"""

  target_name, scenario = datalab.mlalpha.Metadata(metadata_path).get_target_name_and_scenario()
  target_type = 'float_list' if scenario == 'continuous' else 'int64_list'
  content_definitions = \
"""def extract_values((example, prediction)):
  import tensorflow as tf
  tf_example = tf.train.Example()
  tf_example.ParseFromString(example.values()[0])
  feature_map = tf_example.features.feature
  values = {'target': feature_map['%s'].%s.value[0]}
  values.update(prediction)
  return values

OUTPUT_DIR = '%s'
%s
""" % (target_name, target_type, output_dir, _pipeline_definition_code(is_cloud, 'evaluate'))

  content_evaluation = \
"""
eval_features = (pipeline | 'ReadEval' >> io.LoadFeatures('%s'))
trained_model = pipeline | 'LoadModel' >> io.LoadModel('%s')
evaluations = (eval_features | 'Evaluate' >> ml.Evaluate(trained_model) |
    beam.Map('ExtractEvaluationResults', extract_values))
eval_data_sink = beam.io.TextFileSink(os.path.join(OUTPUT_DIR, 'eval'), shard_name_template='')
evaluations | beam.io.textio.WriteToText(os.path.join(OUTPUT_DIR, 'eval'), shard_name_template='')
""" % (preprocessed_eval_data_path, model_dir)

  output_analysis = (output_prediction_name is not None and scenario != 'continuous')
  content_analysis = ''
  if output_analysis:
    if output_score_name is not None:
      score_value = """values['%s'][values['%s']]""" % (output_score_name, output_prediction_name)
    else:
      score_value = '0.0'
      
    content_analysis = \
"""def make_data_for_analysis(values):
  return {
      'target': values['target'],
      'predicted': values['%s'],
      'score': %s,
  }

metadata = pipeline | io.LoadMetadata('%s')
analysis_source = evaluations | beam.Map('CreateAnalysisSource', make_data_for_analysis)
confusion_matrix, precision_recall, logloss = (analysis_source |
    'Analyze Model' >> analysis.AnalyzeModel(metadata))
confusion_matrix_file = os.path.join(OUTPUT_DIR, 'analyze_cm.json')
confusion_matrix_sink = beam.io.TextFileSink(confusion_matrix_file, shard_name_template='')
confusion_matrix | beam.io.Write('WriteConfusionMatrix', confusion_matrix_sink)
""" % (output_prediction_name, score_value, metadata_path)

  content_run = """pipeline.run()"""

  content = \
"""
# header
%s
# imports
%s
# defines
%s
# evaluation
%s
# analysis
%s
# run pipeline
%s
""" % (content_header, content_imports, content_definitions,
       content_evaluation, content_analysis, content_run)

  if output_analysis:
    content += """
# View Confusion Matrix with the following code:
#
# import datalab.mlalpha
# import yaml
# with ml.util._file.open_local_or_gcs(confusion_matrix_file, 'r') as f:
#   data = [yaml.load(line) for line in f.read().rstrip().split('\\n')]
# datalab.mlalpha.ConfusionMatrix([d['predicted'] for d in data],
#                            [d['target'] for d in data],
#                            [d['count'] for d in data]).plot()
"""
  IPython.get_ipython().set_next_input(content, replace=True)


def _evaluate(args, cell):
  if not cell:
    _output_evaluate_template(args['cloud'])
    return

  env = datalab.utils.commands.notebook_environment()
  config = datalab.utils.commands.parse_config(cell, env)
  datalab.utils.commands.validate_config(config,
     ['preprocessed_eval_data_path', 'metadata_path', 'model_dir', 'output_dir'],
     optional_keys=['output_prediction_name', 'output_score_name'])
  command = '%%mlalpha evaluate'
  if args['cloud']:
    command += ' --cloud'
  command += '\n' + cell
  _output_evaluate_code_template(command, args['cloud'], config['preprocessed_eval_data_path'],
      config['metadata_path'], config['model_dir'], config['output_dir'],
      output_prediction_name=config.get('output_prediction_name', None),
      output_score_name=config.get('output_score_name', None))


def _output_dataset_template(name):
  content = """%%mlalpha dataset --name %s
source:
  data1: data_local_or_gcs_path
  data2: data_local_or_gcs_path
featureset: your-featureset-class-name
""" % name
  IPython.get_ipython().set_next_input(content, replace=True)


def _dataset(args, cell):
  if not cell:
    _output_dataset_template(args['name'])
    return
  env = datalab.utils.commands.notebook_environment()
  config = datalab.utils.commands.parse_config(cell, env)
  datalab.utils.commands.validate_config(config, ['source', 'featureset'],
      optional_keys=['format'])
  if config['featureset'] not in env:
    raise Exception('"%s" is not defined.' % config['featureset'])
  featureset_class = env[config['featureset']]
  format = config.get('format', 'csv')
  ds = datalab.mlalpha.DataSet(featureset_class(), config['source'], format=format)
  env[args['name']] = ds


def _module(args, cell):
  if not cell:
    raise Exception('Expect code in cell.')
    return

  env = datalab.utils.commands.notebook_environment()
  if '_ml_modules_' not in env:
    modules = {}
    env['_ml_modules_'] = modules
  modules = env['_ml_modules_']
  modules[args['name']] = cell
  if args['main']:
    env['_ml_modules_main_'] = args['name']


def _package(args, cell):
  env = datalab.utils.commands.notebook_environment()
  if '_ml_modules_' not in env:
    raise Exception('No ml modules defined. Expect modules defined with "%%mlalpha module"')
  package_path = datalab.mlalpha.Packager().package(env['_ml_modules_'], args['name'])
  google.cloud.ml.util._file.create_directory(args['output'])
  dest = os.path.join(args['output'], os.path.basename(package_path))
  google.cloud.ml.util._file.copy_file(package_path, dest)
  os.remove(package_path)
  print 'Package created at %s.' % dest



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Iterator class for iterable cloud lists."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object


class Iterator(object):
  """An iterator implementation that handles paging over a cloud list."""

  def __init__(self, retriever):
    """Initializes an instance of an Iterator.

    Args:
      retriever: a function that can retrieve the next page of items.
    """
    self._page_token = None
    self._first_page = True
    self._retriever = retriever
    self._count = 0

  def __iter__(self):
    """Provides iterator functionality."""
    while self._first_page or (self._page_token is not None):
      items, next_page_token = self._retriever(self._page_token, self._count)

      self._page_token = next_page_token
      self._first_page = False
      self._count += len(items)

      for item in items:
        yield item

  def reset(self):
    """Resets the current iteration."""
    self._page_token = None
    self._first_page = True
    self._count = 0



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Helper functions for %%sql modules."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from past.builtins import basestring
from builtins import object

import shlex

from . import _sql_statement
from . import _utils


# It would be nice to be able to inherit from Python module but AFAICT that is not possible.
# So this just wraps a bunch of static helpers.

class SqlModule(object):
  """ A container for SqlStatements defined together and able to reference each other. """

  @staticmethod
  def _get_sql_args(parser, args=None):
    """ Parse a set of %%sql arguments or get the default value of the arguments.

    Args:
      parser: the argument parser to use.
      args: the argument flags. May be a string or a list. If omitted the empty string is used so
          we can get the default values for the arguments. These are all used to override the
          arg parser. Alternatively args may be a dictionary, in which case it overrides the
          default values from the arg parser.
    Returns:
      A dictionary of argument names and values.
    """
    overrides = None
    if args is None:
      tokens = []
    elif isinstance(args, basestring):
      command_line = ' '.join(args.split('\n'))
      tokens = shlex.split(command_line)
    elif isinstance(args, dict):
      overrides = args
      tokens = []
    else:
      tokens = args

    args = {} if parser is None else vars(parser.parse_args(tokens))
    if overrides:
      args.update(overrides)

    # Don't return any args that are None as we don't want to expand to 'None'
    return {arg: value for arg, value in args.items() if value is not None}

  @staticmethod
  def get_default_query_from_module(module):
    """ Given a %%sql module return the default (last) query for the module.

    Args:
      module: the %%sql module.

    Returns:
      The default query associated with this module.
    """
    return _utils.get_default_query_from_module(module)

  @staticmethod
  def get_sql_statement_with_environment(item, args=None):
    """ Given a SQLStatement, string or module plus command line args or a dictionary,
     return a SqlStatement and final dictionary for variable resolution.

    Args:
      item: a SqlStatement, %%sql module, or string containing a query.
      args: a string of command line arguments or a dictionary of values.

    Returns:
      A SqlStatement for the query or module, plus a dictionary of variable values to use.
    """
    if isinstance(item, basestring):
      item = _sql_statement.SqlStatement(item)
    elif not isinstance(item, _sql_statement.SqlStatement):
      item = SqlModule.get_default_query_from_module(item)
      if not item:
        raise Exception('Expected a SQL statement or module but got %s' % str(item))

    env = {}
    if item.module:
      env.update(item.module.__dict__)
      parser = env.get(_utils._SQL_MODULE_ARGPARSE, None)
      if parser:
        args = SqlModule._get_sql_args(parser, args=args)
      else:
        args = None

    if isinstance(args, dict):
      env.update(args)

    return item, env

  @staticmethod
  def expand(sql, args=None):
    """ Expand a SqlStatement, query string or SqlModule with a set of arguments.

    Args:
      sql: a SqlStatement, %%sql module, or string containing a query.
      args: a string of command line arguments or a dictionary of values. If a string, it is
          passed to the argument parser for the SqlModule associated with the SqlStatement or
          SqlModule. If a dictionary, it is used to override any default arguments from the
          argument parser. If the sql argument is a string then args must be None or a dictionary
          as in this case there is no associated argument parser.
    Returns:
      The expanded SQL, list of referenced scripts, and list of referenced external tables.
    """
    sql, args = SqlModule.get_sql_statement_with_environment(sql, args)
    return _sql_statement.SqlStatement.format(sql._sql, args)





# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Datalab - notebook functionality."""


import httplib2 as _httplib2
import requests as _requests


try:
  import IPython as _IPython
  import IPython.core.magic as _magic
  import IPython.core.interactiveshell as _shell
  from IPython import get_ipython
except ImportError:
  raise Exception('This package requires an IPython notebook installation')

import datalab.context as _context

# Import the modules that do cell magics.
import datalab.bigquery.commands
import datalab.context.commands
import datalab.data.commands
import datalab.mlalpha.commands
import datalab.stackdriver.commands
import datalab.storage.commands
import datalab.utils.commands


_orig_request = _httplib2.Http.request
_orig_init = _requests.Session.__init__
_orig_run_cell_magic = _shell.InteractiveShell.run_cell_magic
_orig_run_line_magic = _shell.InteractiveShell.run_line_magic


def load_ipython_extension(shell):
  """
  Called when the extension is loaded.

  Args:
      shell - (NotebookWebApplication): handle to the Notebook interactive shell instance.
  """

  # Inject our user agent on all requests by monkey-patching a wrapper around httplib2.Http.request.

  def _request(self, uri, method="GET", body=None, headers=None,
               redirections=_httplib2.DEFAULT_MAX_REDIRECTS, connection_type=None):
    if headers is None:
      headers = {}
    headers['user-agent'] = 'GoogleCloudDataLab/1.0'
    return _orig_request(self, uri, method=method, body=body, headers=headers,
                         redirections=redirections, connection_type=connection_type)

  _httplib2.Http.request = _request

  # Similarly for the requests library.

  def _init_session(self):
    _orig_init(self)
    self.headers['User-Agent'] = 'GoogleCloudDataLab/1.0'

  _requests.Session.__init__ = _init_session

  # Be more tolerant with magics. If the user specified a cell magic that doesn't
  # exist and an empty cell body but a line magic with that name exists, run that
  # instead. Conversely, if the user specified a line magic that doesn't exist but
  # a cell magic exists with that name, run the cell magic with an empty body.

  def _run_line_magic(self, magic_name, line):
    fn = self.find_line_magic(magic_name)
    if fn is None:
      cm = self.find_cell_magic(magic_name)
      if cm:
        return _run_cell_magic(self, magic_name, line, None)
    return _orig_run_line_magic(self, magic_name, line)

  def _run_cell_magic(self, magic_name, line, cell):
    if len(cell) == 0 or cell.isspace():
      fn = self.find_line_magic(magic_name)
      if fn:
        return _orig_run_line_magic(self, magic_name, line)
      # IPython will complain if cell is empty string but not if it is None
      cell = None
    return _orig_run_cell_magic(self, magic_name, line, cell)

  _shell.InteractiveShell.run_cell_magic = _run_cell_magic
  _shell.InteractiveShell.run_line_magic = _run_line_magic

  # Define global 'project_id' and 'set_project_id' functions to manage the default project ID. We
  # do this conditionally in a try/catch # to avoid the call to Context.default() when running tests
  # which mock IPython.get_ipython().

  def _get_project_id():
    try:
      return _context.Context.default().project_id
    except Exception:
      return None

  def _set_project_id(project_id):
    context = _context.Context.default()
    context.set_project_id(project_id)

  def _get_bq_dialect():
    return datalab.bigquery.Dialect.default().bq_dialect

  def _set_bq_dialect(bq_dialect):
    datalab.bigquery.Dialect.default().set_bq_dialect(bq_dialect)

  try:
    if 'datalab_project_id' not in _IPython.get_ipython().user_ns:
      _IPython.get_ipython().user_ns['datalab_project_id'] = _get_project_id
      _IPython.get_ipython().user_ns['set_datalab_project_id'] = _set_project_id
      _IPython.get_ipython().user_ns['datalab_bq_dialect'] = _get_bq_dialect
      _IPython.get_ipython().user_ns['set_datalab_bq_dialect'] = _set_bq_dialect
  except TypeError:
    pass


def unload_ipython_extension(shell):
  _shell.InteractiveShell.run_cell_magic = _orig_run_cell_magic
  _shell.InteractiveShell.run_line_magic = _orig_run_line_magic
  _requests.Session.__init__ = _orig_init
  _httplib2.Http.request = _orig_request
  try:
    del _IPython.get_ipython().user_ns['project_id']
    del _IPython.get_ipython().user_ns['set_project_id']
  except Exception:
    pass  # We mock IPython for tests so we need this.
  # TODO(gram): unregister imports/magics/etc.



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements listing projects and setting default project."""

from __future__ import absolute_import
from __future__ import unicode_literals
try:
  import IPython
  import IPython.core.magic
  import IPython.core.display
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import fnmatch

import datalab.utils.commands
import datalab.context


@IPython.core.magic.register_line_cell_magic
def projects(line, cell=None):
  parser = datalab.utils.commands.CommandParser.create('projects')

  list_parser = parser.subcommand('list', 'List available projects.')
  list_parser.add_argument('-f', '--filter',
                           help='Optional wildcard id filter string used to limit the results')
  list_parser.set_defaults(func=_list_line, cell_prohibited=True)

  set_parser = parser.subcommand('set', 'Set the default project.')
  set_parser.add_argument('id', help='The ID of the project to use')
  set_parser.set_defaults(func=_set_line, cell_prohibited=True)

  return datalab.utils.commands.handle_magic_line(line, cell, parser)


def _list_line(args, _):
  # TODO(gram): should we use a paginated table?
  filter_ = args['filter'] if args['filter'] else '*'
  data = [{'id': project.id, 'name': project.name} for project in datalab.context.Projects()
                       if fnmatch.fnmatch(project.id, filter_)]
  return IPython.core.display.HTML(datalab.utils.commands.HtmlBuilder.render_table(data, ['id', 'name']))


def _set_line(args, _):
  id_ = args['id'] if args['id'] else ''
  context = datalab.context.Context.default()
  context.set_project_id(id_)
  datalab.context.Projects.save_default_id(id_)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - IPython HTML display Functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from builtins import range
from past.builtins import basestring
from builtins import object

import time


class Html(object):
  """A helper to enable generating an HTML representation as display data in a notebook.

  This object supports the combination of HTML markup and/or associated JavaScript.
  """

  _div_id_counter = 0

  @staticmethod
  def next_id():
    """ Return an ID containing a reproducible part (counter) and unique part (timestamp). """
    Html._div_id_counter += 1
    return '%d_%d' % (Html._div_id_counter, int(round(time.time() * 100)))

  def __init__(self, markup=None):
    """Initializes an instance of Html.
    """
    self._id = Html.next_id()
    Html._div_id_counter += 1
    self._class_name = ''
    self._markup = markup
    self._dependencies = [('element!hh_%d' % self._id, 'dom')]
    self._script = ''
    self._class = None

  def add_class(self, class_name):
    """Adds a CSS class to be generated on the output HTML.
    """
    self._class = class_name

  def add_dependency(self, path, name):
    """Adds a script dependency to be loaded before any script is executed.
    """
    self._dependencies.append((path, name))

  def add_script(self, script):
    """Adds JavaScript that should be included along-side the HTML.
    """
    self._script = script

  def _repr_html_(self):
    """Generates the HTML representation.
    """
    parts = []
    if self._class:
      parts.append('<div id="hh_%s" class="%s">%s</div>' % (self._id, self._class, self._markup))
    else:
      parts.append('<div id="hh_%s">%s</div>' % (self._id, self._markup))

    if len(self._script) != 0:
      parts.append('<script>')
      parts.append('require([')
      parts.append(','.join(['"%s"' % d[0] for d in self._dependencies]))
      parts.append('], function(')
      parts.append(','.join([d[1] for d in self._dependencies]))
      parts.append(') {')
      parts.append(self._script)
      parts.append('});')
      parts.append('</script>')

    return ''.join(parts)


class HtmlBuilder(object):
  """A set of helpers to build HTML representations of objects.
  """

  def __init__(self):
    """Initializes an instance of an HtmlBuilder.
    """
    self._segments = []

  def _render_objects(self, items, attributes=None, datatype='object'):
    """Renders an HTML table with the specified list of objects.

    Args:
      items: the iterable collection of objects to render.
      attributes: the optional list of properties or keys to render.
      datatype: the type of data; one of 'object' for Python objects, 'dict' for a list
          of dictionaries, or 'chartdata' for Google chart data.
    """
    if not items:
      return

    if datatype == 'dict':
      getter = lambda obj, attribute: obj.get(attribute, None)
    elif datatype == 'chartdata':
      if not attributes:
        attributes = [items['cols'][i]['label'] for i in range(0, len(items['cols']))]
      items = items['rows']
      indices = {attributes[i]: i for i in range(0, len(attributes))}
      getter = lambda obj, attribute: obj['c'][indices[attribute]]['v']
    else:
      getter = lambda obj, attribute: obj.__getattribute__(attribute)

    num_segments = len(self._segments)
    self._segments.append('<table>')

    first = True
    for o in items:
      if first:
        first = False
        if datatype == 'dict' and not attributes:
          attributes = list(o.keys())

        if attributes is not None:
          self._segments.append('<tr>')
          for attr in attributes:
            self._segments.append('<th>%s</th>' % attr)
          self._segments.append('</tr>')

      self._segments.append('<tr>')
      if attributes is None:
        self._segments.append('<td>%s</td>' % HtmlBuilder._format(o))
      else:
        for attr in attributes:
          self._segments.append('<td>%s</td>' % HtmlBuilder._format(getter(o, attr), nbsp=True))
      self._segments.append('</tr>')

    self._segments.append('</table>')
    if first:
      # The table was empty; drop it from the segments.
      self._segments = self._segments[:num_segments]

  def _render_text(self, text, preformatted=False):
    """Renders an HTML formatted text block with the specified text.

    Args:
      text: the text to render
      preformatted: whether the text should be rendered as preformatted
    """
    tag = 'pre' if preformatted else 'div'
    self._segments.append('<%s>%s</%s>' % (tag, HtmlBuilder._format(text), tag))

  def _render_list(self, items, empty='<pre>&lt;empty&gt;</pre>'):
    """Renders an HTML list with the specified list of strings.

    Args:
      items: the iterable collection of objects to render.
      empty: what to render if the list is None or empty.
    """
    if not items or len(items) == 0:
      self._segments.append(empty)
      return
    self._segments.append('<ul>')
    for o in items:
      self._segments.append('<li>')
      self._segments.append(str(o))
      self._segments.append('</li>')
    self._segments.append('</ul>')

  def _to_html(self):
    """Returns the HTML that has been rendered.

    Returns:
      The HTML string that has been built.
    """
    return ''.join(self._segments)

  @staticmethod
  def _format(value, nbsp=False):
    if value is None:
      return '&nbsp;' if nbsp else ''
    elif isinstance(value, basestring):
      return value.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
    else:
      return str(value)

  @staticmethod
  def render_text(text, preformatted=False):
    """Renders an HTML formatted text block with the specified text.

    Args:
      text: the text to render
      preformatted: whether the text should be rendered as preformatted
    Returns:
      The formatted HTML.
    """
    builder = HtmlBuilder()
    builder._render_text(text, preformatted=preformatted)
    return builder._to_html()

  @staticmethod
  def render_table(data, headers=None):
    """ Return a dictionary list formatted as a HTML table.

    Args:
      data: a list of dictionaries, one per row.
      headers: the keys in the dictionary to use as table columns, in order.
    """
    builder = HtmlBuilder()
    builder._render_objects(data, headers, datatype='dict')
    return builder._to_html()

  @staticmethod
  def render_chart_data(data):
    """ Return a dictionary list formatted as a HTML table.

    Args:
      data: data in the form consumed by Google Charts.
    """
    builder = HtmlBuilder()
    builder._render_objects(data, datatype='chartdata')
    return builder._to_html()

  @staticmethod
  def render_list(data):
    """ Return a list formatted as a HTML list.

    Args:
      data: a list of strings.
    """
    builder = HtmlBuilder()
    builder._render_list(data)
    return builder._to_html()
